import datetime
import arrow
from typing import Optional, Union, Callable
from collections import namedtuple
from ..entities import Entity
from .query import Query

optional_time = Optional[Union[datetime.datetime, arrow.Arrow]]

Criterion: namedtuple

class Criteria:
    def by_id(self, *ids: int) -> Query: ...
    def by_location(self, *locations: object) -> Query: ...
    def by_path(self, *paths: str) -> Query: ...
    def by_version(self, *versions: int) -> Query: ...
    def by_name(self, *names: str) -> Query: ...
    def by_component(self, *components: str) -> Query: ...
    def by_status(self, *status: str) -> Query: ...
    def by_shot(self, *shots: str) -> Query: ...
    def by_sequence(self, *sequences: str) -> Query: ...
    def by_task(self, *tasks: str) -> Query: ...
    def by_type(self, *types: str) -> Query: ...
    def by_project(self, *projects: str) -> Query: ...
    def by_metadata(self, **dictionaries: str) -> Query: ...
    def by_description(self, regex: str) -> Query: ...
    def by_comment(self, regex: str) -> Query: ...
    def by_publisher(self, *publishers: str) -> Query: ...
    def by_assetgroup(self, *assetgroups: str) -> Query: ...
    def by_state(self, *states: str) -> Query: ...
    def by_list(self, *lists: str) -> Query: ...
    def by_publish_time(self, start: optional_time = None, end: optional_time = None) -> Query: ...
    def by_status_change_time(self, start: optional_time = None, end: optional_time = None) -> Query: ...
    def by_assignee(self, *assignees: str) -> Query: ...
    def by_lifespan(self, start: optional_time = None, end: optional_time = None) -> Query: ...
    def by_outgoing_link(self, *ids: str) -> Query: ...
    def by_incoming_link(self, *ids: str) -> Query: ...
    def inject(self, filter: str) -> Query: ...

    @staticmethod
    def supported_targets(*supported: Entity) -> Callable: ...


from .base import Entity
from typing import *

class ForwardDeclaration: ...

class RelationshipDeclaration:
    def __init__(self, parent: Union[Type['Entity'], Type[ForwardDeclareCompare]], child: Union[Type['Entity'], Type[ForwardDeclareCompare], str]): ...

class ForwardDeclareCompare(type):
    def __eq__(self, other: Union[Type['Entity'], Type[ForwardDeclaration]]) -> bool: ...

class Asset(ForwardDeclaration): ...
class Scope(ForwardDeclaration): ...
class Metadata(ForwardDeclaration): ...
class TaskTemplateItem(ForwardDeclaration): ...
class ReviewSessionInvitee(ForwardDeclaration): ...
class Bid(ForwardDeclaration): ...
class Milestone(ForwardDeclaration): ...
class SequenceComponent(ForwardDeclaration): ...
class ObjectType(ForwardDeclaration): ...
class ReviewSessionObjectStatus(ForwardDeclaration): ...
class Task(ForwardDeclaration): ...
class Project(ForwardDeclaration): ...
class Conversation(ForwardDeclaration): ...
class AssetVersionLink(ForwardDeclaration): ...
class Timelog(ForwardDeclaration): ...
class Shot(ForwardDeclaration): ...
class ListCustomAttributeValue(ForwardDeclaration): ...
class Disk(ForwardDeclaration): ...
class Event(ForwardDeclaration): ...
class Participant(ForwardDeclaration): ...
class JobComponent(ForwardDeclaration): ...
class NoteCategory(ForwardDeclaration): ...
class State(ForwardDeclaration): ...
class Setting(ForwardDeclaration): ...
class Location(ForwardDeclaration): ...
class Episode(ForwardDeclaration): ...
class CalendarEvent(ForwardDeclaration): ...
class Type(ForwardDeclaration): ...
class ContextCustomAttributeValue(ForwardDeclaration): ...
class Job(ForwardDeclaration): ...
class TypedContextLink(ForwardDeclaration): ...
class Chapter(ForwardDeclaration): ...
class AssetType(ForwardDeclaration): ...
class Timer(ForwardDeclaration): ...
class CustomAttributeGroup(ForwardDeclaration): ...
class ReviewSession(ForwardDeclaration): ...
class AssetVersion(ForwardDeclaration): ...
class AssetBuild(ForwardDeclaration): ...
class UserCustomAttributeValue(ForwardDeclaration): ...
class ContainerComponent(ForwardDeclaration): ...
class Message(ForwardDeclaration): ...
class CustomAttributeValue(ForwardDeclaration): ...
class TypedContext(ForwardDeclaration): ...
class Group(ForwardDeclaration): ...
class Note(ForwardDeclaration): ...
class WorkflowSchema(ForwardDeclaration): ...
class DashboardWidget(ForwardDeclaration): ...
class Epic(ForwardDeclaration): ...
class Status(ForwardDeclaration): ...
class Appointment(ForwardDeclaration): ...
class CalendarEventResource(ForwardDeclaration): ...
class Membership(ForwardDeclaration): ...
class User(ForwardDeclaration): ...
class ProjectSchemaOverride(ForwardDeclaration): ...
class Recipient(ForwardDeclaration): ...
class TypedContextStatusChange(ForwardDeclaration): ...
class AssetVersionCustomAttributeValue(ForwardDeclaration): ...
class ProjectSchema(ForwardDeclaration): ...
class TaskTemplate(ForwardDeclaration): ...
class EntitySetting(ForwardDeclaration): ...
class ListCategory(ForwardDeclaration): ...
class StatusChange(ForwardDeclaration): ...
class Feed(ForwardDeclaration): ...
class Priority(ForwardDeclaration): ...
class AssetCustomAttributeValue(ForwardDeclaration): ...
class NoteComponent(ForwardDeclaration): ...
class TaskTypeSchema(ForwardDeclaration): ...
class TypedContextList(ForwardDeclaration): ...
class AssetVersionStatusChange(ForwardDeclaration): ...
class UserSecurityRole(ForwardDeclaration): ...
class FileComponent(ForwardDeclaration): ...
class AssetGroup(ForwardDeclaration): ...
class Folder(ForwardDeclaration): ...
class Schema(ForwardDeclaration): ...
class CustomAttributeType(ForwardDeclaration): ...
class Component(ForwardDeclaration): ...
class SchemaType(ForwardDeclaration): ...
class SchemaStatus(ForwardDeclaration): ...
class Resource(ForwardDeclaration): ...
class SecurityRole(ForwardDeclaration): ...
class List(ForwardDeclaration): ...
class ReviewSessionObject(ForwardDeclaration): ...
class CustomAttributeConfiguration(ForwardDeclaration): ...
class Context(ForwardDeclaration): ...
class AssetVersionList(ForwardDeclaration): ...
class Sequence(ForwardDeclaration): ...
class ComponentLocation(ForwardDeclaration): ...
class Dashboard(ForwardDeclaration): ...



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import inspect
import importlib
import logging
import re
from collections import (
    OrderedDict,
    defaultdict,
    namedtuple,
)
from copy import copy
import sys

from ftrack_api.symbol import NOT_SET

from .declarations import *

from .schematypes import (
    CUSTOM_ATTRIBUTE_TYPE_COMPATIBILITY,
    AttributeInfo
)
from ..query import (
    Criterion,
    Criteria,
    utils
)
from ..configuration import (
    LOGGING_NAMESPACE,
    ALLOWED_FOR_DELETION_RESOLVER,
    WARN_ON_INJECT
)


LOG = logging.getLogger("{}.entities".format(LOGGING_NAMESPACE))

_RELATIONSHIPS_CACHE = {}


def TargetRelation(relation="", collection=False):
    """TargetRelation function that "looks" like a class (due to its CamelCase naming on purpose)"""
    return namedtuple(TargetRelation.__name__, ['relation', 'collection'])(relation, collection)


class Relationship(dict):
    """ A simple helper class to check for equality
    """
    parent = None

    def __init__(self):
        super(Relationship, self).__init__()

        self._entity_type_name = None
        self.session = None
        self.schema = None

        try:
            self._entity_type_name = inspect.stack()[1][3]
        except IndexError:
            LOG.warning("Unable to identify Entity Type.", exc_info=True)

    def _infer_from_session_schema(self, session):
        def _to_class(classname):
            return getattr(sys.modules[__name__], classname, None)

        # TODO: check for actual available types
        if self._entity_type_name and self._entity_type_name != "_EntityBase":
            entity_data = session.parsed_relationships.relationships.get(self._entity_type_name)
            if entity_data:
                LOG.debug("Inferring relationship data for {}".format(self._entity_type_name))

                for entity, relation in entity_data.get("non_collection").items():
                    self.add(_to_class(entity), ".".join(relation))
                for entity, relation in entity_data.get("collection").items():
                    self.add(_to_class(entity), ".".join(relation), collection=True)
            else:
                LOG.warning("Couldn't find entry for '{}' in relationships cache.".format(self._entity_type_name))

    @property
    def _cache_key(self):
        server_url, schema_name, entity_type_name = None, None, None
        try:
            server_url = self.session.server_url
            schema_name = self.schema["name"]
            entity_type_name = self._entity_type_name
        except (AttributeError, KeyError):
            pass

        return server_url, schema_name, entity_type_name

    @property
    def global_cache(self):
        return _RELATIONSHIPS_CACHE

    def __repr__(self):
        return str(self)

    def __str__(self):
        key = self._cache_key
        return str(_RELATIONSHIPS_CACHE.get(key, {}))

    def _validate_access(self):
        _error_msg_template = "No {} was set for the current Relationship object. " \
            "Relationship needs to be called with session and schema instance before " \
            "relations can be retrieved."

        # as we do have to read from a global cache we need to assume a session and schema have been defined
        # for our current instance so we can look up previously parsed relationship information from the global
        # cache
        if self.session is None:
            raise AssertionError(_error_msg_template.format("session"))
        if self.schema is None:
            raise AssertionError(_error_msg_template.format("schema"))

    def __getitem__(self, item):
        self._validate_access()

        from_cache = _RELATIONSHIPS_CACHE.get(self._cache_key, {})

        for key, value in from_cache.items():
            if key and (key == item):
                return value
        return super(Relationship, self).__getitem__(item)

    def __setitem__(self, item, value):
        self._validate_access()
        key = self._cache_key

        from_cache = _RELATIONSHIPS_CACHE.get(key, {})
        if not from_cache:
            _RELATIONSHIPS_CACHE[key] = from_cache
        from_cache[item] = value

    def __call__(self, session, schema, entity_type=None):
        # TODO: how to deal with race condition in case queries will run in threads for the same server/session

        # whenever being called store the information about the used session and used schema
        self.session = session
        self.schema = schema
        if entity_type:
            self._entity_type_name = entity_type.__name__

        # build a key for the cached based on used schema name and session server url
        key = self._cache_key

        # on initial request infer all relationships from database and custom schema
        if key not in _RELATIONSHIPS_CACHE:
            # always do a copy of ourselves that will be added to the cache as a Relationship instance
            # is currently being used as a class attribute
            self._infer_from_session_schema(session)

            # TODO: warn if schema doesn't have entities set
            collection_attributes = session.parsed_relationships.array_attributes
            for target_entity_name, relation in schema.get("entities", {}).get(self._entity_type_name, {}).items():
                collection = None

                if not isinstance(relation, list):
                    _relation = [relation]
                else:
                    _relation = relation

                for some_relation in _relation:
                    attribute_tokens = some_relation.split(".")
                    leaf_attribute = some_relation.split(".")[-1]

                    if len(attribute_tokens) > 1:
                        _collection = leaf_attribute in collection_attributes
                    else:
                        _collection = bool(re.match(r"^(children|ancestors)(\[\w+\])?", attribute_tokens[0]))

                    if collection is None:
                        collection = _collection
                    if collection != _collection:
                        raise AttributeError(
                            "Given relationships '{}' attribute array state is ambigious.".format(
                                relation
                            )
                        )

                self.add(globals()[target_entity_name], relation, collection)

    def add(self, target, relation, collection=False):
        """ Adds a new relationship
        """
        self[target] = TargetRelation(relation, collection)

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default


class EmptyCollection(object):
    """ Special collection which is used instead of an EntityColleciton with
    zero Entities. This acts as an Opiontal which will forward any attribute
    getters by always returning itself. This reduces the need for constant
    conditional checks (null checks) and makes the code simpler
    """
    __slots__ = ["_entity", "depth", "source", "_current_iter_index", "_session"]

    def __init__(self, _type=None, source=None, depth=1, session=None):
        self._entity = _type
        self.depth = depth
        self.source = source
        self._current_iter_index = 0
        self._session = session

    def __call__(self, *args, **kwargs):
        return self

    def __contains__(self, item):
        return False

    def __copy__(self):
        return EmptyCollection(
            _type=self._entity,
            depth=self.depth + 1,
            source=self.source,
            session=self._session
        )

    def __eq__(self, other):
        equal = False
        if isinstance(other, EmptyCollection):
            equal = True
            for slot in EmptyCollection.__slots__:
                equal = equal and getattr(self, slot) == getattr(other, slot)
        return equal

    def __getattr__(self, item):
        current_depth = self.depth
        collection_copy = copy(self)
        # in case we use methods like .values() or .from_entities() etc
        # we need to take care to not increment the depth, as it doesn represent
        # an attribute/relation access
        if item in [_ for _ in dir(EntityCollection) if not re.match(r"__\w+__|children", _)]:
            collection_copy.depth = current_depth
        return collection_copy

    def __getitem__(self, item):
        # TEST: Maybe we should return somthing else when accessing POD
        #  attributes so we can avoid doing this:
        #  names.append(Query(Asset).by_name("test").get_all().name or [])

        if inspect.isclass(item) and issubclass(item, Entity):
            self._entity = item()

        return self

    def __iter__(self):
        self._current_iter_index = 0
        return self

    def __len__(self):
        return 0

    def __next__(self):
        if self._current_iter_index == 0:
            self._current_iter_index += 1
            return copy(self)
        else:
            raise StopIteration

    # python 2 compatibility
    def next(self):
        return self.__next__()

    def __nonzero__(self):
        return False

    def __repr__(self):
        return "EmptyCollection[{}]".format(self._entity)

    @property
    def entity_type(self):
        return self._entity

    def create(self, **kwargs):
        assert self.depth == 1, \
            """
            This attribute does not exist yet. The parent attribute also did not exist.
            Create can only be called on existing attributes OR if at least the parent
            exists.
            """
        entitycollection = EntityCollection._make_empty(self._entity.__class__, self._session)
        return entitycollection.create(empty=self, **kwargs)

    def create_batch(self, *attributes):
        """

        Args:
            *attribute ():

        Returns:

        """

        collections = []

        for _attributes in attributes:
            collections.append(self.create(**_attributes))

        return EmptyCollection(_type=self._entity).union(*collections)

    def union(self, *collections):
        if len(collections) == 1:
            return collections[0]
        
        first = collections[0]
        others = [_ for _ in collections[1:] if _]
        first_type = first.entity_type
        for other_type in (_.entity_type for _ in others if _):
            assert first_type == other_type, "Can't union collections with different types. {} != {}."\
                .format(first_type, other_type)
        
        return first.union(*others)


class EntityCollection(object):
    """ An immutable sorted set of unique Entities of the same entity type.
    """
    MEMBERS = [
        "_entity",
        "_entities",
        "_query",
        "query",
        "_schema_types_map",
        "_current_iter_index",
        "_session",
        "_source"
    ]

    def __init__(self, _cls=None, entities={}, session=None):
        super(EntityCollection, self).__init__()
        self._entity = _cls()

        self._entities = entities
        self._query = None
        self._current_iter_index = []
        self._session = session

        # similar to the source attribute on EmptyCollection
        # we sometimes need to keep track of what produced this
        # current collection
        self._source = (None, None)

        self._schema_types_map = {
            "string": (str,),
            "number": (float,),
            "boolean": (bool,),
            "integer": (int,),
            "mapped_array": (dict,)
        }

    def __contains__(self, other):
        if isinstance(other, str):
            return other in self.keys()
        elif isinstance(other, Entity):
            return other in self.values()
        elif isinstance(other, EntityCollection):
            return all([_ in self.keys() for _ in other.keys()])

    def __eq__(self, other):
        if isinstance(other, EntityCollection):
            return [_.ftrack_entity for _ in self.values()] == \
                   [_.ftrack_entity for _ in other.values()]
        else:
            return False

    def __getattr__(self, item):
        # TODO: clean after we have tests for this functionality
        if self and item in list(self.values())[0].ftrack_entity.keys():
            entities = []
            values = []
            entity_type = None
            for entity in self.values():
                value = entity[item]
                # resolve entities in collections
                if value.__class__.__name__ == "Collection":
                    for _ in value:
                        # wrap ftrack entity
                        if hasattr(_, "entity_type"):
                            entities.append(Entity.from_entity_type(str(_.entity_type), ftrack_entity=_))
                        else:
                            values.append(_)
                    if not entities and not values and not entity_type:
                        entity_type = Entity.from_entity_type(
                            name=self._get_attribute_compatibility_types(item).types[0].__name__
                        )
                elif value.__class__.__name__ == "KeyValueMappedCollectionProxy":
                    values.append(dict(entity.ftrack_entity["metadata"].items()))

                # wrap ftrack entity
                elif hasattr(value, "entity_type"):
                    entities.append(Entity.from_entity_type(str(value.entity_type), ftrack_entity=value))
                else:
                    values.append(value)

            if entities:
                return self.from_entities(entities, source=(item, self))
            elif entity_type:
                return EmptyCollection(_type=entity_type, source=self, session=self._session)
            return values

        elif item == "get":
            return self._get_relatives

        # handle custom attributes
        elif item.startswith("custom_"):
            attr = item.replace("custom_", "")
            values = []
            for entity in self.values():
                values.append(entity.ftrack_entity["custom_attributes"].get(attr))
            return values

        # handle collection.<type>
        elif item in globals():
            # if not already parsed and cached let's do it now
            self._entity.relationship(
                session=self.query.session,
                schema=self.query.schema,
                entity_type=self._entity.__class__
            )
            if self._entity.relationship.get(globals()[item]):
                relation = self._entity.relationship.get(globals()[item]).relation

                result = None

                if not isinstance(relation, list):
                    relation = [relation]

                for some_relation in relation:
                    current = None
                    # tokenize the attribute chain
                    for token in some_relation.split("."):
                        _filter = None
                        # and also identify filters like with `parent[Shot]`
                        _ = re.findall(r"\w+", token)
                        if len(_) == 2:
                            token, _filter = _
                        elif len(_) > 2:
                            raise ValueError("Unable to identify attribute and filter within token '{}'.".format(_))

                        _collection = self if current is None else current
                        current = getattr(_collection, token)

                        if NOT_SET in current:
                            self._entity.log.warning(
                                "Was the attribute you're trying to access projected? "
                                "A symbol.NOT_SET found on '{}' access. This would resolve to '{}'.".format(
                                    item,
                                    ", ".join(relation)
                                )
                            )

                        # TODO: check if the entity we need for filtering is in the globals
                        if _filter:
                            current = current[globals()[_filter]]

                    if result is None:
                        result = current
                    else:
                        result = result.union(current)

                return result

        else:
            raise AttributeError("Attribute {} does not exist.".format(item))

    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            # no one needs to know that internally we store the id as the key!
            return self.from_entities(list(self.values())[item])
        elif isinstance(item, str):
            return self.from_entities(self._entities[item])
        elif issubclass(item, Entity):
            return self.resolve_subtype(item)
        # TODO: would there be a scenarios outside our __getattr__ usecase
        #  where this would be a direct ForwardDeclaration subclass?
        elif issubclass(item, ForwardDeclaration):
            return self.resolve_subtype(item)

    def __hash__(self):
        return hash(tuple(self.keys()))

    def __iter__(self):
        self._current_iter_index.append(0)
        return self

    def __len__(self):
        return len(self.keys())

    # TODO: how to cleanup if we have a break statement?
    def __next__(self):
        if self._current_iter_index[-1] >= len(self.keys()):
            del self._current_iter_index[-1]
            raise StopIteration
        else:
            self._current_iter_index[-1] += 1
            return self[self._current_iter_index[-1] - 1]

    # python 2 compatibility
    def next(self):
        return self.__next__()

    def __repr__(self):
        """ Formats self like EntityCollection[EntityType]{NumberOfEntities}.

        Example:
            EntityType[AssetVersion]{23}

        Returns:
            str: Formatted string representing the EntityCollection.
        """
        return "{}[{}]{{{}}}".format(self.__class__.__name__, self._entity, len(self))

    def __setattr__(self, key, value):
        # TODO: setting uses_versions and used_in_versions will not update bidirectionally
        #  Can we solve this in a generic way?
        # should we update this so we have the correct data locally?
        if key in self.MEMBERS:
            super(EntityCollection, self).__setattr__(key, value)
        else:
            attribute_value = getattr(self, key, None)
            if attribute_value is not None:
                # If the attribute access produces a collection
                # we potentially can not go by key as the attribute
                # access can be a type based shortcut.
                # Therefore we need to resolve the collections produced
                # via the full attribute chain and retain the last attribute
                # as our key.
                if not isinstance(attribute_value, list):
                    key, collection = attribute_value._source
                else:
                    collection = self

                compatible, reason = collection._is_attribute_compatible_with_value(key, value)
                if not compatible:
                    raise TypeError(reason)

                if isinstance(value, (list, tuple, EntityCollection)) \
                        and len(collection) == 1 \
                        and list(collection._entities.values())[0].ftrack_entity[key].__class__.__name__ == "Collection":
                    if isinstance(value, EntityCollection):
                        list(collection.values())[0].ftrack_entity[key] = [_.ftrack_entity for _ in value.values()]
                else:
                    for idx, entity in enumerate(collection.values()):
                        if key.startswith("custom_"):
                            if isinstance(value, (list, tuple, EntityCollection)):
                                entity.ftrack_entity["custom_attributes"][key.replace("custom_", "")] = value[idx]
                            else:
                                entity.ftrack_entity["custom_attributes"][key.replace("custom_", "")] = value
                        elif entity.ftrack_entity[key].__class__.__name__ == "KeyValueMappedCollectionProxy":
                            _should_capture = False
                            if isinstance(value, (list, tuple)):
                                if any(_ not in  value[idx].keys() for _ in entity.ftrack_entity[key]):
                                    _should_capture = True
                                for _key, _value in value[idx].items():
                                    entity.ftrack_entity[key][_key] = _value
                            else:
                                if not len(value):
                                    entity.ftrack_entity[key] = value
                                else:
                                    for _key, _value in value.items():
                                        entity.ftrack_entity[key][_key] = _value
                        else:
                            if isinstance(value, EntityCollection):
                                entity.ftrack_entity[key] = list(value.values())[idx].ftrack_entity
                            elif isinstance(value, (list, tuple)):
                                entity.ftrack_entity[key] = value[idx]
                            else:
                                entity.ftrack_entity[key] = value

            else:
                raise AttributeError("Attribute {} does not exist.".format(key))

    def __nonzero__(self):
        return len(self) > 0

    def _init_type_map(self, type_def):
        """ builds a dictionary with EmptyCollections, based on the given type definitions

        Notes:
              This is primarily useful for resolving TypedContext subclasses
        Args:
            type_def (TypesBase subclass): availavle types from schema

        Returns:
            dict where the key is the available type and the value and EmptyCollection of the same type
        """
        # Initialize the dictionary with EmptyCollections of correct type so
        # we can create TypedContexts for object_types that have not been
        # created in the pertinent context
        type_map = {}
        for _type in type_def.types:
            empty = EmptyCollection(
                _type=Entity.from_entity_type(str(_type)),
                source=self,
                session=self._session
            )
            type_map[_type] = empty
        return type_map

    def resolve_subtype(self, entity_type):
        """ filters entities by matching the entity_type returns a collection of the given type including the matches

        Args:
            entity_type (Entity): a given entity class that must be a subclass of the type of the current collection

        Returns:
            EntityCollection[entity_type] or EmptyCollection[entity_type]
        """
        if not issubclass(entity_type, ForwardDeclaration):
            # this can only work if the given entity type is a subclass of our currently
            # used collection type, like TypedContext -> AssetBuild|Shot|Sequence... or Component -> FileComponent
            # TODO: re-think the necessity of checking for the actual base type
            #  Enforcing a type coercion before filtering via entity type makes the code more clear, but is
            #  not strictly needed for correct filtering.
            assert issubclass(entity_type, self.entity_type)
        else:
            entity_type = getattr(importlib.import_module("..entities", __name__), entity_type.__name__)

        matched_types = self.filter(lambda x: list(x.values())[0].ftrack_entity.entity_type == entity_type.__name__)
        return self.from_entities(
            [
                Entity.from_entity_type(entity_type.__name__, ftrack_entity=_.ftrack_entity) for _ in
                matched_types.values() or []  # fall back to empty list in case we already have an EmptyCollection
            ],
            type_override=entity_type,
            source=getattr(matched_types, "source", None)
        )

    @property
    def children(self):
        """ get all children

        Returns:
            EntityCollection[TypedContext]
        """
        # TODO: deal with this when it becomes a problem ;)
        #  Other entities like Context or Group have children too. When we need to
        #  access those we have to implement the functionality properly here.
        # if not isinstance(self._entity, TypedContext):
        #     getattr(self, "children")

        default_projections = ["children.{}".format(_) for _ in self._entity.projections]
        self._simple_children_fetch("children.object_type.name", *default_projections)

        children_entities = []

        for entity in self.values():
            if entity["children"]:
                for child in entity["children"]:
                    children_entities.append(
                        Entity.from_entity_type("TypedContext", ftrack_entity=child)
                    )

        return self.from_entities(children_entities, source=self)

    def query_children(self, projections=None, session=None):
        """Queries and returns the children of the collection. Only supported for
        entities that have a "children" relation

        Args:
            projections (`list` of `str`): Projections to fetch from the children
            session (Session): Optional session to use.

        """
        projections_ = ["children.{}".format(_) for _ in self._entity.projections]
        projections_ += ["children.{}".format(x) for x in (projections or [])]
        projections_.append("children.object_type.name")
        query = self.as_query(use_ids=True)
        query.projections = list(set(projections_).union(query.projections))  # update not override
        return query.get_all(session=session or self._session).children

    # TODO: move to avoid circular import
    @staticmethod
    def _make_empty(entity_class, session):
        # UGLY AS HELL!!!!!
        entitycollection = EntityCollection(_cls=entity_class, entities=OrderedDict())
        # this special import is neccessary to avoid a circular import
        adhoc_type = getattr(importlib.import_module("...query", __name__), "Query")
        entitycollection.query = adhoc_type(entity=entity_class, session=session)
        entitycollection.query.valid = False
        entitycollection._session = session
        return entitycollection

    @property
    def entity_type(self):
        return self._entity.__class__

    def link_inputs(self, collection):
        """ Links compatible Entities as inputs to the current collection.

        Args:
            collection (EntityCollection): Collection of entities to link as inputs

        Returns:
            self
        """
        self.fetch_attributes("incoming_links.from_id")
        if len(collection) != 0:
            for entity in self:
                links = entity.incoming_links
                for _id in collection.id:
                    if _id not in links.from_id:
                        entity.incoming_links.create(from_id=_id, to_id=entity.id[0])
        else:
            LOG.warning("The collection you want to link is empty.")

        return self

    def link_outputs(self, collection):
        """ Links compatible Entities as outputs to the current collection.

        Args:
            collection (EntityCollection): Collection of entities to link as outputs

        Returns:
            self
        """
        self.fetch_attributes("outgoing_links.to_id")
        if len(collection) != 0:
            for entity in self:
                links = entity.outgoing_links
                for _id in collection.id:
                    if _id not in links.to_id:
                        entity.outgoing_links.create(from_id=entity.id[0], to_id=_id)
        else:
            LOG.warning("The collection you want to link is empty.")

        return self

    def unlink_inputs(self, collection):
        """ Deletes all incoming links between collection and self.

        Args:
            collection (EntityCollection): Collection of entities to unlink as inputs

        Returns:
            EntityColleciton of remaining input entities.
        """
        self.fetch_attributes("incoming_links.from_id")
        if len(collection) != 0:
            for entity in self:
                links = entity.incoming_links
                for _id in collection.id:
                    links.filter(lambda x: str(_id) == str(x.from_id[0])).delete()
        else:
            LOG.warning("The collection you want to unlink is empty.")

        return self

    def unlink_outputs(self, collection):
        """ Deletes all outgoing links between self and collection.

        Args:
            collection (EntityCollection): Collection of entities to unlink as outputs.

        Returns:
            EntityColleciton of remaining output entities.
        """
        self.fetch_attributes("outgoing_links.to_id")
        if len(collection) != 0:
            for entity in self:
                links = entity.outgoing_links
                for _id in collection.id:
                    links.filter(lambda x: str(_id) == str(x.to_id[0])).delete()
        else:
            LOG.warning("The collection you want to unlink is empty.")

        return self

    def _is_attribute_compatible_with_value(self, attribute_name, value):
        attribute_info = self._get_attribute_compatibility_types(attribute_name)
        value_info = self._get_value_type(value)
        compatible, reason = True, ""

        # TODO: make sure we compare the actual types and not just the names
        # TEST: What happens if we naively do this. Adjust unittests to check!
        if value_info.types not in [_ for _ in attribute_info.types]:
            compatible, reason = False, \
                (
                    "The given value '{}' does not have the correct type "
                    "for the receiver attribute '{}'. {} != {}".format(
                        value,
                        attribute_name,
                        value_info.types,
                        attribute_info.types
                    )
                )
            return compatible, reason

        if isinstance(value, (EntityCollection, tuple, list)):
            _reason = (
                "When setting an attribute on a receiver collection, "
                "we expect the given value to have the same amount of elements as our collection has entities "
                "or in case it is an iterable to have the same length."
            )
            if attribute_info.array:
                if len(self) > 1:
                    # we want to assign a value to multiple entities
                    if len(self) != len(value):
                        compatible, reason = False, _reason
            elif len(self) != len(value):
                compatible, reason = False, _reason
        elif len(self) > 1:
            # value is not a Collection/Array, but we contain multiple entities
            compatible, reason = False, \
                "The given value is not a collection or array, but the receiver contains more than a single entity."

        return compatible, reason

    @staticmethod
    def _get_value_type(value):
        if isinstance(value, (EntityCollection, EmptyCollection)):
            return AttributeInfo(
                types=type(value._entity),
                array=True
            )
        elif isinstance(value, (tuple, list)):
            return AttributeInfo(
                types=type(value[0]),
                array=True
            )
        else:
            return AttributeInfo(types=type(value))

    def _get_attribute_compatibility_types(self, attribute_name):
        """ Returns an AttributeInfo object that encodes which values can
        potentially be assigned to self.[attribute_name]
            Args:
                 attribute_name (str): The name of the attribute we want to check.

            Returns:
                AttributeInfo: An AttributeInfo object that encodes the values that
                self.[attribute_name] can consume.
        """
        if attribute_name.startswith("custom_"):
            attribute_name = attribute_name.replace("custom_", "")
            attribute_info = getattr(CUSTOM_ATTRIBUTE_TYPE_COMPATIBILITY, attribute_name)
            return attribute_info

        entity_schema = None

        for schema in self._session.schemas:
            if schema["id"] == self._entity.__class__.__name__:
                entity_schema = schema

        property_schema = entity_schema["properties"][attribute_name]

        attribute_info = AttributeInfo()

        if property_schema.get("$ref"):
            attribute_info.types = (globals()[property_schema["$ref"]],)
        elif property_schema.get("key_value_attributes"):
            attribute_info.types = self._schema_types_map[property_schema["type"]]
        elif property_schema.get("items"):
            attribute_info.array = True
            attribute_info.types = (globals()[property_schema["items"]["$ref"]],)
        elif property_schema.get("type"):
            attribute_info.types = self._schema_types_map[property_schema["type"]]
        else:
            raise AttributeError("Attribute '{}' not found in schema '{}'.".format(attribute_name, entity_schema["id"]))

        return attribute_info

    def keys(self):
        return self._entities.keys()

    def values(self):
        return self._entities.values()

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, instance):
        """ Sets the query to a copy of the given query instance.
        """
        self._query = copy(instance)

    def from_entities(self, entities, type_override=None, source=None):
        """ helper to create an EntityCollection from given entities

        Args:
            entities (Entity or list): Entity subclass or list of Entity subclasses
            source (EntityCollection or tuple(str, EntityCollection)): store what source produced the new collection
            to preserve the source, from where the resulting EmptyCollection was generated from.
            type_override: if given it will use the override as the type for the generated collection
                and not determine it from the first entity automatically

        Returns:
            EntityCollection
        """
        if not entities:
            if type_override:
                type_override = type_override()
            return EmptyCollection(session=self._session, _type=type_override or self._entity, source=source)

        if isinstance(entities, Entity):
            # TEST: list() constructor too slow, why???
            entities = [entities]

        _entities = OrderedDict(
            [
                (_.id,  _ if not type_override else Entity.from_entity_type(type_override.__name__, _.ftrack_entity))
                for _ in entities
            ]
        )

        # This is just a temporary "solution".
        # We hit the situation that our first entity type lookup results in being a `FileComponent`, whereas
        # other entities can be of type `Component`, which is an unlucky case that lead to fundamental problem
        # when fetching. We need to ensure we don't give subclasses the precedence here!
        # This sorting acts as a very cheap way and we assume that a subclass of a specific type will always
        # include its base type in the name.
        _entity_types = sorted(
            list(set([_.__class__ for _ in _entities.values()])),
            key=lambda x: len(x.__name__.split(".")[-1])
        )

        # TODO: handle ForwardDeclaration

        entities = EntityCollection(
            _cls=type_override or _entity_types[0],
            entities=_entities,
            session=self._session
        )

        entities.query = self.query
        entities.query.valid = False
        entities._source = source

        return entities

    def group(self, predicate):
        """ Returns a dictionary with keys given by the predicate. All entities
        from the original collection will be mapped to their corresponding key.

        Examples:
            >>> groups = asset_versions.group(lambda x: x.asset.name)
            >>> print(groups)

            {
                "gurke_aa": EntityCollection[AssetVersion],
                "banane_ab": EntityCollection[AssetVersion]
            }

        Args:
            predicate (callable): callable object that returns an attribute value

        Returns:
            dict: dictionary including the given predicate result as key and an
            EntityCollection as value
        """
        temp = defaultdict(list)
        groups = {}

        for entity in self:
            temp[predicate(entity)].append(list(entity.values())[0])

        for group, entities in temp.items():
            groups[group] = self.from_entities(entities)

        return groups

    def fold(self, startvalue, predicate):
        """ Accumulates the value starting with an initial value and applying
        an operation from the first to the last element in a collection.

        Args:
            startvalue (Any): it can be anything!!!!
            predicate (callable): callable object that returns an attribute value

        Returns:
            Type(startvalue): Value based on the folding.

        """
        for entitycollection in self:
            startvalue = predicate(startvalue, entitycollection)
        return startvalue

    def map(self, predicate):
        """ Returns a list of results that is obtained by running the predicate
        with each Entity as it's argument and returning the result.
        """
        return map(predicate, self)

    def group_and_map(self, group_predicate, map_predicate):
        """ Runs a group_by first and then runs map on all the Collection in
        the resulting dictionary values.
        """
        groups = self.group(group_predicate)

        mapped_groups = {}
        for key, value in groups.items():
            mapped_groups[key] = map_predicate(value)

        return mapped_groups

    def count(self, predicate):
        """ Returns the number of elements for which the predicate returns True

        Args:
            predicate (callable): callable object that returns an attribute value

        Returns:

        """
        return len(self.filter(predicate))

    def max(self, predicate):
        """ Returns the first element yielding the largest value of the given function.

        Args:
            predicate (callable): callable object that returns an attribute value

        Returns:
            EntityCollection: A new EntityCollection containing the maximum Element

        """
        return self.sort(predicate)[-1]

    def min(self, predicate):
        """ Returns the first element yielding the smallest value of the given function.

        Args:
            predicate (callable): callable object that returns an attribute value

        Returns:
            EntityCollection: A new EntityCollection containing the minimum Element

        """
        # If multiple entities have the same min value, we follow python's min implementation
        # by returning the first occurence of this value. In contrast, the max implementation
        # returns the last occurence.
        _sorted = self.sort(predicate)
        min_value = predicate(_sorted[0])
        return _sorted.filter(lambda x: predicate(x) == min_value)[0]

    def sort(self, predicate, reverse=False):
        """ Returns a duplicate of the original collection sorted by the given predicate.

        Args:
            predicate (callable): callable that returns a comparable object / value
            reverse (bool): reverse the sorting order

        Returns:
            EntityCollection: A duplicate of the original collection sorted by the given predicate.

        """

        entity_collections = sorted(self, key=lambda x: max(predicate(x)), reverse=reverse)

        if len(entity_collections) > 1:
            return entity_collections[0].union(*entity_collections[1:])
        else:
            return entity_collections[0]

    def filter(self, predicate):
        """ Filters the EntityCollection based on a given predicate.

        The predicate is run on each Entity and returns a boolean (or is coerced
        to a boolean). If True, the Entity will be added to the resulting
        collection, if False it will not be added.

        Args:
            predicate (callable): most likely a lambda referring to an entity attribute

        Returns:
            EntityCollection: filtered collection

        """

        filtered = []

        for entitycollection in self:
            if predicate(entitycollection):
                for entity in entitycollection.values():
                    filtered.append(entity)

        if filtered:
            return self.from_entities(filtered)
        else:
            return EmptyCollection(_type=self._entity, source=self._get_parent(self), session=self._session)

    def apply(self, predicate, attribute_name=None):
        """Applies a predicate function to each entity in the collection and
        assigns the generated value to the specified attribute.
        If no attribute name is provided, the value is directly assigned to the calling collection.

        Examples:
            >>> some_collection.apply(lambda c: c.another_attr[0] + "_edited", "some_attr")
            >>> assetversion_collection.Task.Status.apply(status_collection)

        Args:
            predicate (callable): A callable function that receives a single entity collection.
                The return value of the function will be applied to the specified attribute.
            attribute_name (optional: str): The name of the attribute to which the generated value will be assigned
                or None

        Returns:
            EntityCollection: the updated collection

        """
        if not attribute_name:
            attribute_name, collection = self._source
        else:
            collection = self

        for entitycollection in collection:
            setattr(entitycollection, attribute_name, predicate(entitycollection))

        return getattr(collection, attribute_name)

    @staticmethod
    def _get_parent(entitycollection):
        parent_relation = entitycollection._entity.relationship.parent or "parent"
        parents = getattr(entitycollection, parent_relation)
        if not isinstance(parents, (EntityCollection, EmptyCollection)):
            entitycollection.fetch_attributes(parent_relation)
            parents = getattr(entitycollection, parent_relation)

        msg = "Cannot determine singular parent. `{}` ".format(entitycollection)
        if not parents:
            LOG.warning(msg + "has no parents.")
        elif len(parents) > 1:
            LOG.warning(msg + "has multiple parents.")

        return parents

    def partition(self, predicate):
        """ Splits original collection into tuple of EntityCollections,
        where the first EntityCollection contains elements for which the predicate returned True,
        while the second EntityCollection contains elements for which the predicate returned False.

        Args:
            predicate (callable):

        Returns:
            tuple: A tuple including the positive results first and negative results second

        """
        positive, negative = [], []

        for entitycollection in self:
            if predicate(entitycollection):
                for entity in entitycollection.values():
                    positive.append(entity)
            else:
                for entity in entitycollection.values():
                    negative.append(entity)

        return self.from_entities(positive), self.from_entities(negative)

    def union(self, *collections):
        """ Returns an EntityCollection with elements belonging to the current or the given collection or possibly both

        Args:
            collections (EntityCollection): EntityCollection object of same type as the current one

        Returns:
            EntityCollection:

        """
        entities = list(self.values())
        for collection in collections:
            if not collection:
                continue  # avoid merging EmptyCollections
            self._validate_collection_type(collection)
            for entity in collection:
                if list(entity.values())[0] not in entities:
                    entities.append(list(entity.values())[0])

        return self.from_entities(entities)

    def intersection(self, *collections):
        """ Returns an EntityCollection with elements belonging to the current and the given collection

        Args:
            collections (EntityCollection): EntityCollection object of same type as the current one

        Returns:
            EntityCollection:

        """
        for collection in collections:
            self._validate_collection_type(collection)

        intersection = []
        for _id in self.keys():
            if all(_id in collection.keys() for collection in collections):
                intersection.append(_id)

        return self.filter(lambda x: x.id[0] in intersection)

    # TODO: Mathematically this is a binary operation and needs to be applied in a specific order.
    #  The current implementation could be renamed to 'unique' or similar.
    # FIX: The current implementation of symmetric difference does not return
    #  the expected result.
    def symmetric_difference(self, *collections):
        """ Returns an EntityCollection with elements belonging possibly to the
        current and the given collection, but never to both.

        Args:
            *collections (EntityCollection): EntityCollection objects of same
             type as the current one

        Returns:
            EntityCollection
        """
        for collection in collections:
            self._validate_collection_type(collection)

        counter = defaultdict(int)

        all_collections = [self] + list(collections)
        all_entities = self.union(*collections)

        for collection in all_collections:
            for entity in collection.values():
                counter[entity] += 1

        difference = [entity for entity in all_entities.values() if counter[entity] == 1]

        return self.from_entities(difference)

    def difference(self, *collections):
        """ Generates the difference between self and the given collections.

        This essentially acts as a "subtraction" operation.

        Returns:
            EntityCollection
        """

        ids = self.keys()

        for collection in collections:
            ids = [_id for _id in ids if _id not in collection.keys()]

        return self.filter(lambda x: x.id[0] in ids)

    def _validate_collection_type(self, collection):
        """ Compares a given collection type with the collection type of self.

        Raises an error if the collections are not compatible i.e. we can't
        store their entities in the same collection.
        """

        # TODO: allow set operations if the source collection is a parent type of the given collection type
        #  This is useful if you have a EntityCollection[TypeContext] you want to operate on
        # is_valid = True
        # parent_types = ["TypedContext", "List"]
        #
        # if self.entity_type != collection.entity_type:
        #     is_valid = False
        #     if (self.values() and collection.values()) and list(self.values())[0].__class__.__name__ in parent_types:
        #         if issubclass(list(collection.values())[0].__class__, list(self.values())[0].__class__):
        #             is_valid = True
        #
        # if not is_valid:
        #    ...

        if self.entity_type != collection.entity_type:
            raise EntityCollectionOperationError(
                "Only EntityCollections of same type can be used for Set operations. "
                "Expected {}, got {}".format(repr(self), repr(collection))
            )

    def as_query(self, use_ids=False, with_projections=False):
        # TODO(high): To support dynamic ftrack entities we need to have the
        #  possibility to use different attributes than id in this case.
        #  Potentially use ftrack_api.session.populate for fetching!.
        # TEST: How do we know that we have a dynamicftrackentity:
        #  Generalize handling of non fixed primary keys.
        if use_ids or not self.query.valid:
            self.query.criteria = [
                Criterion(
                    name="by_id",
                    args=tuple(self.keys()),
                    kwargs={},
                    target=None,
                    filter=self._entity.by_id
                )
            ]
            if not with_projections:
                self.query.projections = []
            self.query.entity_type = self._entity

        return self.query

    def fetch_attributes(self, *attributes):
        """ Fetches the given attributes from ftrack by doing another query.
        This allows access if auto population is disabled

        Args:
            *attributes (str): attributes (can be scoped)
        """

        def _deepen(iterables):
            nested = {}
            for iterable in iterables:
                temp = nested
                for value in iterable:
                    if not temp.get(value):
                        temp[value] = {}
                    temp = temp[value]
            return nested

        def _flatten(iterable):
            def __flatten(iterable, path=""):
                flat = {}
                if len(iterable.keys()) == 1:
                    # we are on a non branching level
                    key = list(iterable.keys())[0]
                    # append to the existing path as we don't need to branch
                    path += "." + key if path else key  # only insert dot in existing path
                    # get the downstream dict recursively
                    if key.endswith("]"):
                        _ = __flatten(iterable[key], "")
                    else:
                        _ = __flatten(iterable[key], path)
                    # if the downstream level will not branch either, we move the dict
                    # up one level. this is also done if we split due to a type
                    # specialization e.g. [Shot]
                    if len(_.keys()) == 1 and not key.endswith("]"):
                        return _
                    # otherwise we insert a level for a branch
                    flat[path] = _
                    return flat
                elif len(iterable.keys()) > 1:
                    # we are on a branching level
                    for key, value in iterable.items():
                        if len(value.keys()) == 1:
                            # if the child will not branch, we will reset the
                            # hierarchy to not accumulate duplicate levels
                            path = ""
                        _ = __flatten(value, path)
                        if len(_.keys()) == 1:
                            # if the childs are not branching we extract the values
                            # and set them with the correct key
                            flat[key + "." + list(_.keys())[0]] = list(_.values())[0]
                        else:
                            # otherwise we will create a new branch
                            flat[key] = _
                    return flat
                return flat

            flat = __flatten(iterable)
            return flat

        def _get_attribute(entity, attributes):
            attribute_tokens = attributes.split(".")

            parent = entity
            for token in attribute_tokens:
                if token.endswith("]"):
                    token, _type = token.split("[")
                entity = getattr(parent, token)
                # TODO: this is not ideal, we shouldn't run into this situation
                #  so we probably have to improve the _flatten function
                if isinstance(entity, list) and NOT_SET in entity:
                    _fetch(parent, {token: {}})
                    entity = getattr(parent, token)

                parent = entity

            return entity

        def _fetch(entities, attributes):
            # TODO(nice to have): Can we potentially do a fetch on an EmptyCollection by using the 'source' attribute?
            #  This is only possible if we fetch into the direction of the source attribute.
            if isinstance(entities, EmptyCollection):
                LOG.info("Encountered an `EmptyCollection` while fetching attributes. Can't continue fetching.")
            else:
                query = entities.as_query(use_ids=True)
                if isinstance(attributes, dict):
                    _bases = query.entity_type.__class__.__bases__

                    # There is the chance of getting a collection containing different types
                    # of the same base type when constructed via _get_attribute(), as we don't handle filtering
                    # like parent[AssetBuild]. This leads to the situation that our query would only fetch the
                    # requested attributes for type of the first entity in the collection. Therefore we have to
                    # ensure we use the base type when fetching.
                    if _bases and _bases[0].__name__ in ["TypedContext"]:
                        query.entity_type = _bases[0]()

                    query.projections = [_.split("[")[0] for _ in attributes.keys()]
                    query.get_all()

                    for key, value in attributes.items():
                        entity = _get_attribute(entities, key)
                        if not isinstance(entity, list):
                            _fetch(entity, value)
                else:
                    query.projections = attributes
                    query.get_all(session=self._session)

        # lets resolve all RelationshipDeclarations first
        _attributes = []
        for attribute in attributes:
            if isinstance(attribute, RelationshipDeclaration):
                _attributes.extend(attribute.resolve_path_for(self.entity_type, self.query.session, self.query.schema))
            else:
                _attributes.append(attribute)

        attributes = [attribute.split(".") for attribute in _attributes]
        deep = _deepen(attributes)
        flat = _flatten(deep)
        # attribute_hierarchy = _flatten(_deepen(attributes))
        _fetch(self, flat)

        return self

    def _simple_children_fetch(self, *attributes):
        # this is needed as a special case to avoid an infinite recursion due to
        # fetch_attributes calling getattr and getatrr calling fetch_attributes
        # (wheh accessing children)
        query = self.as_query(use_ids=True)
        # TODO(high): Reimplement the conditional fetching of non-existing projections.
        # Update the projections on the query object accordingly.
        # only perform a query if really necessary
        if set(attributes).union(query.projections) != set(query.projections):
            query.projections = list(set(attributes).union(query.projections))  # update not override
            query.get_all()
        else:
            LOG.info("All requested attributes already set.")

    def commit(self):
        self._session.commit()

    def _get_relatives(self, relative_type, **kwargs):
        """ Get the relative entity based on the relationship.

        Args:
            relative_type (Entity): subclass of Entity
            **kwargs ():

        Returns:
            Any: whatever the attributes holds, most likely another EntityCollection
        """

        relationship = self.query.entity_type.relationship.get(relative_type, default=TargetRelation()).relation
        if not relationship:
            raise ValueError("Unknown relationship for relative '{}'".format(relative_type))

        if not isinstance(relationship, list):
            relation_projections = [relationship]
        else:
            relation_projections = relationship

        outside_projections = kwargs.get("projections", [])
        constructed_projections = []

        for outside_projection in outside_projections:
            for relation_projection in relation_projections:
                constructed_projections.append("{}.{}".format(relation_projection, outside_projection))

        if not outside_projections:
            constructed_projections.extend(relation_projections)

        self.fetch_attributes(*constructed_projections)

        return getattr(self, relative_type.__name__)

    def _prepare_note_entity(self, ftrack_note_entity, kwargs, parent_source):
        """ preparation of a node entity

        This shall ensure we don't have to pass author and recipients for `notes.create()`.

        Args:
            ftrack_note_entity (entity): the ftrack api entity instance

        Returns:

        """
        # TODO: this definitely needs cleanup
        # make the Query class available
        _Query = getattr(importlib.import_module("...query", __name__), "Query")

        # get the actual entity implemantions not the ForwardDeclare classes
        _Recipient = getattr(importlib.import_module("..entities", __name__), "Recipient")
        _User = getattr(importlib.import_module("..entities", __name__), "User")

        if not ftrack_note_entity.get("author"):
            ftrack_note_entity["author"] = list(_Query(_User, session=self._session).by_name(
                self._session.api_user).get_one().values())[0].ftrack_entity

        recipients = ftrack_note_entity.get("recipients", [])
        if not recipients:
            resource_ids = kwargs.get("recipient_resource_ids", [])
            recipients = _Query(_Recipient, session=self._session).get_first()
            if not resource_ids:
                resource_ids = [ftrack_note_entity["author"]["id"]]

            for resource_id in resource_ids:
                recipient = recipients.create(
                    note_id=ftrack_note_entity["id"],
                    resource_id=resource_id
                )
                ftrack_note_entity["recipients"].append(list(recipient.values())[0].ftrack_entity)

        # The parent source is a sibling collection, reuse it's parent attrs
        if parent_source.entity_type.__name__ == "Note":
            parent_source.fetch_attributes("parent_type", "parent_id")
            # TODO: improve error message when context is ambiguous
            assert len(set(parent_source.parent_id)) == 1, "Ambiguous context. Multiple parents found."
            ftrack_note_entity["parent_id"] = parent_source.parent_id[0]
            ftrack_note_entity["parent_type"] = parent_source.parent_type[0]

        # The parent source is a the actual parent of the node, use it's attributes to
        # create the first note.
        else:
            ftrack_note_entity["parent_id"] = parent_source.id[0]
            ftrack_note_entity["parent_type"] = parent_source._entity.__class__.__name__

    def create(self, **kwargs):
        """
        Creates a new Entity based on the current context.

        The context is defined by self as well as the potential parent attribute.
        If a parent is set in the relationships of an Entity, we will use that,
        otherwise we simply assume an attribute name of "parent".
        We can also ignore the requirement of a parent by providing the
        no_parent keyword argument (the value does not matter). This is useful
        for entities that don't require (or have) a parent, like Project.

        Args:
            **kwargs: A dictionary of data that is required to initialize the
            Entity properly.

        Returns:
            An EntityCollection with the created Entity.

        """

        # TODO: Handle creation of dynamic ftrack entities.
        # TODO: this can fail if we have not specified a parent relation and
        #  the default relation does not exist. In this case we would fetch a
        #  non existing attribute which will fail. Raise a clear assertion which
        #  describes the exact circumstances of the failure.
        #  At the same time it is required to have the 'no_parent' set.

        # checking for the "empty" kwarg enables us to create new Entities
        # without requiring preexisting siblings.
        if "empty" in kwargs:
            this = kwargs.get("empty")
            parent = this.source
            parent_relation = kwargs.get("parent_relation", None) or this._entity.relationship.parent or "parent"
            del kwargs["empty"]
        else:
            this = self
            parent_relation = kwargs.get("parent_relation", None) or this._entity.relationship.parent or "parent"

            if parent_relation in list(this.values())[0].ftrack_entity.keys():
                if not all(getattr(this, parent_relation)):
                    this.fetch_attributes(parent_relation)

                parent = getattr(this, parent_relation)
            else:
                kwargs["no_parent"] = True

        # pre-process the keyword arguments for the corresponding entity type
        kwargs["session"] = self._session
        kwargs = this._entity.pre_create(**kwargs)

        if "no_parent" not in kwargs.keys():
            # TODO: improve error message when context is ambiguous
            assert len(parent) == 1, "Ambiguous context. Multiple parents found."

            kwargs.update(
                {parent_relation: list(parent.values())[0].ftrack_entity}
            )

        assert isinstance(kwargs, dict), "pre_create method returned something that is not a dict"

        entity_class_name = this._entity.__class__.__name__
        ftrack_entity = self._session.create(entity_class_name, data=kwargs)

        # specific handling for anything that can't be easily achieved within the pre-create method
        if entity_class_name == "Note":
            # When the entity we have doesn't have any notes, we can use "parent" as the
            # source of information, this will be the collection from which the notes are
            # being created.
            if isinstance(parent, EntityCollection):
                note_parent_source = parent
            # If the collection already has some notes, "this" will be the collection
            # for the existing siblings, we can then use those to extract the information.
            else:
                note_parent_source = this
            self._prepare_note_entity(ftrack_entity, kwargs, parent_source=note_parent_source)

        entity = this._entity.__class__(_cls=this._entity.__class__, ftrack_entity=ftrack_entity)

        return self.from_entities([entity])

    def create_batch(self, *attributes):
        """ Calls "create" in a loop to create multiple entities and store all
        of them in an EntityCollection.


        Args:
            *attributes (dict): Multiple dictionaries with key-value-pairs that
            will be used as the keyword arguments for the "create" call.

        Returns:
            An EntityCollection with all the Entities that have been created.
        """

        collections = []

        for _attributes in attributes:
            collections.append(self.create(**_attributes))

        return EmptyCollection(_type=self._entity, session=self._session).union(*collections)

    def delete(self):
        """ Deletes all Entities in self from ftrack.

        Returns:
            The SESSION object so we can call commit() directly.
        """
        type_name = self.entity_type.__name__
        if not ALLOWED_FOR_DELETION_RESOLVER(session=self._session, type_name=type_name):
            raise AssertionError(
                "Current entity type '{}' for server '{}' is not allowed for deletion. ".format(
                    type_name,
                    self._session.server_url
                )
            )

        for entity in self.values():
            ftrack_entity = entity.ftrack_entity
            LOG.info("Deleting Entity '{}'".format(ftrack_entity))
            self._session.delete(ftrack_entity)

        return self._session


class _EntityBase(object, metaclass=ForwardDeclareCompare):

    relationship = Relationship()
    projections = ["id"]
    _ftrack_entity = None
    log = None

    def __new__(cls, *args, **kwargs):
        """ make it possible to swap the Entity class with the given class

        Args:
            *args ():
            **kwargs ():

        Returns:

        """
        if kwargs.get("_cls"):
            cls = kwargs["_cls"]
            cls.log = logging.getLogger("{}.entities.{}".format(cls.__name__, LOGGING_NAMESPACE))
            del kwargs["_cls"]

        if cls and args and isinstance(args[0], (EntityCollection, EmptyCollection)):
            source_collection = args[0]

            if source_collection:
                _cls = source_collection.entity_type
            else:
                _cls = source_collection.entity_type.__class__

            assert issubclass(_cls, cls), \
                "Can't coerce `{source}` to `{target}`, because `{source}` is not a subtype of `{target}`.".format(
                    source=_cls.__name__,
                    target=cls.__name__,
                )

            _source = getattr(source_collection, "source", None)

            target_collection = source_collection.from_entities(
                source_collection.values() or [],
                type_override=cls,
                source=_source
            ) or \
            EmptyCollection(
                session=source_collection._session,
                _type=cls(),
                source=_source
            )  # TODO:  shall we implement from_entities on EmptyCollection?

            return target_collection

        return super(_EntityBase, cls).__new__(cls)

    def __init__(self, _cls=None, ftrack_entity=None, **kwargs):
        self.ftrack_entity = ftrack_entity
        if not _cls:
            self.log = logging.getLogger(
                "{}.entities.{}".format(self.__class__.__name__, LOGGING_NAMESPACE)
            )

    def __getitem__(self, item):
        return self.ftrack_entity.get(item)

    def __getattr__(self, item):
        value = None
        if item in self.ftrack_entity.keys():
            value = self.ftrack_entity[item]
            if hasattr(value, "entity_type"):
                return Entity.from_entity_type(str(value.entity_type), ftrack_entity=value)
        return value

    def __str__(self):
        return self.__class__.__name__

    def __hash__(self):
        return hash(self.ftrack_entity["id"])

    def __eq__(self, other):
        # TEST: Figure out how to avoid this behaviour and ditch this workaround.
        #  weirdly there is a possibility that this is executed with classes as
        #  self (which by definition should NEVER be class) and other
        #  (for example when comparing EmptyCollections)
        #  see TestRelativeTerminators for an example
        #  Additionally, this should actually work, as we've overridden __eq__
        #  in the metaclass which should take care of class comparisons
        if inspect.isclass(self) and inspect.isclass(other):
            return self == other
        elif isinstance(self, Entity) and isinstance(other, Entity):
            return self.ftrack_entity == other.ftrack_entity
        else:
            return False

    def __cmp__(self, other):
        if self == other:
            return 0
        else:
            return 1

    def pre_create(self, **kwargs):
        return kwargs

    @property
    def ftrack_entity(self):
        return self._ftrack_entity

    @ftrack_entity.setter
    def ftrack_entity(self, instance):
        self._ftrack_entity = instance

    @classmethod
    def from_entity_type(cls, name, ftrack_entity=None):
        entity_cls = getattr(importlib.import_module("..entities", __name__), str(name))
        if not issubclass(entity_cls, Entity):
            raise NotImplementedError("The Entity of type \"{}\" has not been implemented yet.".format(entity_cls.__name__))
        return cls(_cls=entity_cls, ftrack_entity=ftrack_entity)

    # These methods are only here so the stubs builder can automatically add them.
    def get(self, projections=None, limit=None, offset=0, order="ascending", order_by=None): pass
    def get_first(self, projections=None): pass
    def get_one(self, projections=None): pass
    def get_all(self, projections=None): pass
    def get_inputs(self, projections=None): pass
    def get_outputs(self, projections=None): pass

    def create(self, **kwargs): pass
    def create_batch(self, *attributes): pass
    def delete(self): pass

    # although there is no logical reason to have multiple filters
    # this is how the query consolidates its criteria
    # having multiple filters allows us the inject criterion multiple times
    # Example:
    #   Query(Task).inject("parent.parent.name is 'library').by_name(Project, "Foobar").inject("parent.status.name is 'Approved').get_all()
    def inject(self, target, *filter):
        if any(re.search(r"^or\s*", _, flags=re.IGNORECASE) for _ in filter):
            raise ValueError(
                "OR relationships are not supported across multiple criteria or filter.\n"
                "You can use an 'or' relationship within a single filter string though."
            )

        # sanitize partial query
        sanitized_filters = [re.sub(r"^(and|where)\s*", "", _, flags=re.IGNORECASE) for _ in filter]
        partial_query = " and ".join(sanitized_filters)
        if WARN_ON_INJECT:
            self.log.warning("`inject` criteria was used with partial query '{}'".format(partial_query))

        return partial_query


class Entity(_EntityBase):

    @Criteria.supported_targets(_EntityBase)
    def by_id(self, target, *ids):
        target = target or TargetRelation()
        return utils.build_partial_query(target, ids, "id")

    @Criteria.supported_targets(_EntityBase)
    def by_metadata(self, target, *dictionaries):
        query = "("
        for _dict in dictionaries:
            query += "("
            for key, value in _dict.items():
                if target:
                    query += "{}.".format(target.relation)
                query += "metadata any (key like \"{}\" and value like \"{}\") and ".format(key, value)
            query = query[:-5] + ") or "
        query = query[:-4] + ")"
        return query


class EntityCollectionOperationError(ValueError):
    pass





#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import itertools


# TODO: This need class/function docs
class TreeData:
    def __init__(self, type=None, parent_attr=None, parent=None):
        self.type = type
        self.parent_attr = parent_attr
        self.parent = parent


class TreeItem:
    def __init__(self, data=TreeData()):
        self.children = []
        self.data = data

    def get_parent(self):
        return self.data.parent


class RelationshipsParser(object):

    def __init__(self, ftrack_session=None):
        self._session = ftrack_session
        self._exclude_types = set()
        self._tmp_exclude_types = set()
        self._ref_key = "refs"
        self._entities = {}
        self._tmp_relations = {}
        self._include_types = set()
        self._attributes_blacklist = set()
        self._root_item = TreeItem()
        self._project_relations = {}
        self._relationships = {}
        self._all_array_types = {}

    def parse_session_schemas(self):
        for ftrack_entity in self._session.schemas:
            entity_id = ftrack_entity.get("id")
            self._entities[entity_id] = {}
            tmp_refs = {}
            tmp_array_refs = {}
            for prop, prop_data in ftrack_entity["properties"].items():
                items = prop_data.get("items")
                ref = prop_data.get("$ref", "")
                if ref:
                    tmp_refs[prop] = ref
                if items:
                    ref = prop_data.get("items").get("$ref", "")
                    if ref:
                        tmp_array_refs[prop] = ref
            self._entities[entity_id]["refs"] = tmp_refs
            self._entities[entity_id]["array_refs"] = tmp_array_refs

    def create_entity_network(self, ftrack_entity):
        self._tmp_exclude_types = set(self._exclude_types)
        # entity type must be excluded, as finding it in one of its children would lead to an endless recursion
        self._tmp_exclude_types.add(ftrack_entity)
        return self.recurse_entity_connections(ftrack_entity, ancestors=set())

    def recurse_entity_connections(self, ftrack_entity, ancestors=None):
        # ancestors are the entities, in sequence, linking to the current entity
        # the top entity of course has none
        if ancestors is None:
            ancestors = set()
        # refs - reference type attributes on the current entity
        # type - current entity's type
        network = {"refs": {}, "type": ftrack_entity}
        # return otherwise recursion will be endless
        if ftrack_entity in ancestors:
            return network

        # get all array or non-array reference attributes of the current entity
        refs = self._entities.get(ftrack_entity).get(self._ref_key)
        if refs:
            # first iteration - validate props
            valid_props = {}
            for prop_name, prop_type in refs.items():
                if prop_name not in self._attributes_blacklist and prop_type not in self._tmp_exclude_types and prop_type != ftrack_entity:
                    valid_props[prop_name] = prop_type
            # add valid prop types to exclude list for deeper recursions
            self._tmp_exclude_types.update(valid_props.values())

            # second iteration, recurse all valid props
            for prop_name, prop_type in valid_props.items():
                if prop_type in self._entities:
                    ancestors.add(ftrack_entity)
                    network["refs"][prop_name] = self.recurse_entity_connections(prop_type, ancestors)
        return network

    def extract_entity_relations(self, network):
        self._tmp_relations = {}
        self.recurse_entity_relations(network)
        return self._tmp_relations

    def recurse_entity_relations(self, network, path=None):
        if path is None:
            path = []
        if "refs" in network:
            for attr_name, attr_data in network.get("refs").items():
                prop_type = attr_data.get("type")
                if prop_type not in self._tmp_relations:
                    tmp_path = list(path)
                    tmp_path.append(attr_name)
                    self._tmp_relations[prop_type] = tmp_path
                    self.recurse_entity_relations(attr_data, tmp_path)

    def extract_all_entity_relations(self):
        for entity_name in self._entities:
            self._relationships[entity_name] = {}

            # create network of non-array references
            self.use_array_refs = False
            asset_network = self.create_entity_network(entity_name)

            # extract non-collection relations
            self._relationships[entity_name]["non_collection"] = self.extract_entity_relations(asset_network)

            # create network of array references
            self.use_array_refs = True
            asset_network = self.create_entity_network(entity_name)

            # extract collection relations
            self._relationships[entity_name]["collection"] = self.extract_entity_relations(asset_network)

    def parse_project_structure(self, project):
        root_data = TreeData(project.entity_type, None)
        self._root_item = TreeItem(root_data)
        self._project_relations = {}
        self.recurse_project_structure(self._root_item, project)

    def recurse_project_structure(self, structure_item=TreeItem(), ftrack_entity=None):
        entity_type = structure_item.data.type
        if entity_type in self._entities:
            refs = self._entities.get(entity_type).get("array_refs")
            for ref in refs:
                if ref in self._attributes_blacklist:
                    continue
                collection = ftrack_entity.get(ref)
                if collection and collection.__class__.__name__ == "Collection":
                    for child in collection:
                        if child.get("parent") == ftrack_entity:
                            child_data = TreeData(child.entity_type, "parent", structure_item)
                            child_item = TreeItem(child_data)
                            structure_item.children.append(child_item)
                            self.recurse_project_structure(child_item, child)
                        elif child.entity_type in self._include_types:
                            tmp_collection_parent_attr = self._relationships.get(child.entity_type).get("non_collection").get(entity_type)
                            if tmp_collection_parent_attr:
                                collection_parent = tmp_collection_parent_attr[-1]
                                if child.get(collection_parent) == ftrack_entity:
                                    child_data = TreeData(child.entity_type, collection_parent, structure_item)
                                    child_item = TreeItem(child_data)
                                    structure_item.children.append(child_item)
                                    self.recurse_project_structure(child_item, child)

    def print_project_structure(self, tree_item=TreeItem(), offset=""):
        print(offset, "Item:", tree_item.data.type, tree_item.data.parent_attr)
        offset = offset + "  "
        for child in tree_item.children:
            self.print_project_structure(child, offset)

    def extract_project_relations(self):
        self._project_relations = {}
        self.walk_tree_down(self._root_item)
        for entity_name, entity_data in self._project_relations.items():
            self._relationships[entity_name]["parent_attr"] = entity_data.get("parent_attr")
            entity_relations = entity_data.get("parent_paths")
            self._relationships[entity_name]["project"] = self.optimize_project_relations(entity_name, entity_relations)

    def walk_tree_down(self, tree_item=TreeItem()):
        for child in tree_item.children:
            child_data = child.data
            if child_data.type not in self._project_relations:
                self._project_relations[child_data.type] = {}
                self._project_relations[child_data.type]["parent_attr"] = child_data.parent_attr
                self._project_relations[child_data.type]["parent_paths"] = []
            tmp_paths = []
            self.walk_tree_up(tmp_paths, child)
            for path in tmp_paths:
                if path not in self._project_relations[child_data.type]["parent_paths"]:
                    self._project_relations[child_data.type]["parent_paths"].append(path)
            self.walk_tree_down(child)

    def walk_tree_up(self, tmp_paths, tree_item=TreeItem(), path=None):
        if path is None:
            path = []
        # top reached
        if tree_item.data.type not in self._project_relations:
            return
        parent = tree_item.get_parent()
        if parent:
            path.append({"type": parent.data.type, "attr": tree_item.data.parent_attr})
            tmp_paths.append(list(path))
            self.walk_tree_up(tmp_paths, parent, path)

    def optimize_project_relations(self, entity, relations):
        if not relations:
            return
        optimized_relations = []
        for relation_data in relations:
            relation = relation_data[-1].get("type")
            if len(relation_data) > 1: # can"t get shorter than 1
                original_path_length = len(relation_data)
                shortest_relation = []
                shorter_path_found = False
                for rel in relation_data[:original_path_length-1]: # now search up the relation to find even shorter ones
                    non_collection_relation = self._relationships.get(rel.get("type")).get("non_collection")
                    shortest_relation.append(rel)
                    if relation in non_collection_relation:
                        alternative_relation = non_collection_relation.get(relation)
                        if len(alternative_relation) <= len(shortest_relation) and len(alternative_relation) == 1:
                            for alt_relation in alternative_relation:
                                shortest_relation.append({"type": relation, "attr": alt_relation})
                                shorter_path_found = True
                            break
                if shorter_path_found:
                    optimized_relations.append(shortest_relation)
                    continue
            optimized_relations.append(relation_data)
        return optimized_relations

    @property
    def exclude_types(self):
        return self._exclude_types

    @exclude_types.setter
    def exclude_types(self, exclude_types):
        self._exclude_types = exclude_types

    @property
    def use_array_refs(self):
        return self._ref_key == "array_refs"

    @use_array_refs.setter
    def use_array_refs(self, use_array_refs):
        if use_array_refs:
            self._ref_key = "array_refs"
        else:
            self._ref_key = "refs"

    @property
    def include_types(self):
        return self._include_types

    @include_types.setter
    def include_types(self, include_types):
        self._include_types = include_types

    @property
    def attributes_blacklist(self):
        return self._attributes_blacklist

    @attributes_blacklist.setter
    def attributes_blacklist(self, attributes_blacklist):
        self._attributes_blacklist = attributes_blacklist

    @property
    def entities(self):
        return self._entities

    @property
    def project_relations(self):
        return self._project_relations

    @property
    def relationships(self):
        return self._relationships

    @property
    def array_attributes(self):
        array_refs = list(set(itertools.chain.from_iterable(_["array_refs"].keys() for _ in self.entities.values())))
        array_refs.sort()
        return array_refs


#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import inspect
import logging
import re

from collections import namedtuple

import wrapt

from ..configuration import LOGGING_NAMESPACE

LOG = logging.getLogger("{}.criteria".format(LOGGING_NAMESPACE))


class Criterion(object):
    """ a helper class to store and resolve a query criterion"""
    def __init__(self, name="", args=(), kwargs={}, target=None, filter=lambda *args, **kwargs: "", negate=False):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.target = target
        self.filter = filter
        self.negate = negate

    def resolve(self):
        partial_query = self.filter(self.target, *self.args, **self.kwargs)
        if self.negate:
            partial_query = re.sub("\s+(in|like)\s+", r" not_\1 ", partial_query)
            partial_query = re.sub("\s+is\s+", " is_not ", partial_query)

        return partial_query

    # stay compatible - previously we used a namedtuple as Criterion object
    def as_dict(self):
        return {
            "name": self.name,
            "args": self.args,
            "kwargs": self.kwargs,
            "target": self.target,
            "filter": self.filter,
            "negate": self.negate
        }


class Criteria():
    """ stubs for filtering (aka query criteria)

    This is used to allow proper autocompletion within (some) IDE(s).
    """

    class InvalidArgumentsError(ValueError): pass
    class CriterionNotImplementedError(NotImplementedError): pass


    def by_id(self, *ids):
        raise NotImplementedError

    def by_location(self, *locations):
        raise NotImplementedError

    def by_resource_identifier(self, *resource_identifiers):
        raise NotImplementedError

    def by_version(self, *versions):
        raise NotImplementedError

    def by_name(self, *names):
        raise NotImplementedError

    def by_type(self, *types):
        raise NotImplementedError

    def by_metadata(self, **dictionaries):
        raise NotImplementedError

    def by_description(self, regex):
        raise NotImplementedError

    def by_comment(self, regex):
        raise NotImplementedError

    def by_publisher(self, *publishers):
        raise NotImplementedError

    def by_status(self, *statuses):
        raise NotImplementedError

    def by_state(self, *states):
        raise NotImplementedError

    def by_list(self, *lists):
        raise NotImplementedError

    def by_publish_time(self, start=None, end=None):
        raise NotImplementedError

    def by_status_change_time(self, start=None, end=None):
        raise NotImplementedError

    def by_assignee(self, *assignees):
        raise NotImplementedError

    def by_lifespan(self, start=None, end=None):
        raise NotImplementedError

    def by_outgoing_link(self, *ids):
        raise NotImplementedError

    def by_incoming_link(self, *ids):
        raise NotImplementedError

    def by_component_location(self, *component_locations):
        raise NotImplementedError

    def by_file_type(self, *file_types):
        raise NotImplementedError

    def by_system_type(self, *system_types):
        raise NotImplementedError

    def by_size(self, *sizes):
        raise NotImplementedError

    def by_active_state(self, *states):
        raise NotImplementedError

    def by_action(self, *actions):
        raise NotImplementedError

    def by_data(self, *datas):
        raise NotImplementedError

    def by_publish_state(self, publish_state):
        raise NotImplementedError

    def inject(self, filter):
        raise NotImplementedError

    def by_creation_date(self, *dates):
        raise NotImplementedError

    def by_finish_date(self, *dates):
        raise NotImplementedError

    @staticmethod
    def supported_targets(*supported):
        """ Wraps the filter methods and validates the filter targets against
        the given supported targets. Raises an error if a given target is
        not supported.

        Args:
            *supported (Entity): arbitrary number of supported entities

        Returns:
            func (filter method): original filter method
        """

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            _supported_class_names = [_.__name__ for _ in supported]

            if args:
                if (not [_ for _ in _supported_class_names if _ in ["_EntityBase", "Entity"]]) and inspect.isclass(args[0]):
                    if not ((args[0] in supported) or bool([_ for _ in supported if args[0].__bases__[0] == _])):
                        raise Criteria.InvalidArgumentsError(
                            "You provided an unsupported target {} in `.{}()`. {}.".format(
                                args[0].__name__,
                                wrapped.__name__,
                                "No targets supported" if not supported else "Supported targets: " + ", ".join(
                                    [_.__name__ for _ in supported]
                                )
                            )
                        )

            # Propagate target as TargetRelation object
            if args[0]:
                args = [instance.relationship[args[0]]] + list(args[1:])
            return wrapped(*args, **kwargs)

        return wrapper


#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import inspect
import logging
import os
import re
import sys

import ftrack_api


DEFAULT_STUBS_PATH = os.path.dirname(__file__)
STUBS_HEADER = "\nimport typing\n"

LOG = logging.getLogger("trackteroid.stubs")

SESSION = ftrack_api.Session(auto_populate=True, auto_connect_event_hub=False)


def make_file(path, mode=0o777, default_content="", overwrite=False):
    """
    Create a file (if it does not yet exist).
    Args:
        path (str):
        mode (octal, optional): file mode
        default_content (str): optional - the default content to put into the file if it get created here
        overwrite (bool): whether the file should be overwritten, if it already exists

    Returns:
        bool - True if file exists or could be created. False if not.
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)
    if os.path.exists(os.path.dirname(path)):
        try:
            if not os.path.exists(path) or overwrite:
                with open(path, 'w') as file_write:
                    file_write.write(default_content)
                os.chmod(path, mode)
            return True
        except (OSError, AttributeError, TypeError):
            LOG.error("Failed to create file '%s'" % path, exc_info=True)
            return False


class StubClassBuilder(object):
    """ A simply and limited helper to build class stubs """

    CLASS_TEMPLATE = "class {class_name}{class_bases}:\n"
    CLASS_MEMBER_TEMPLATE = "    {member_name}: {type} \n"
    INSTANCE_ATTRIBUTE_TEMPLATE = "        self.{attribute_name}: {type} \n"
    CLASS_ATTRIBUTE_TEMPLATE = "    {attribute_name}: {type} \n"
    METHOD_TEMPLATE = "    def {method_name}({arguments}{keyword_arguments}) -> {return_type}:{ellipsis} \n"

    TYPE_MAP = {
        "array": "{ref} = {ref}",
        "string": "str = str()",
        "number": "float = float()",
        "boolean": "bool = bool()",
        "integer": "int = int()",
        "mapped_array": "typing.List = [{ref}]",
        "variable": "typing.Any = None",
        "any": "{ref} = {ref}"
    }

    def __init__(self, name, bases=()):
        self._content = {
            "header": "",
            "class_members": [],
            "constructor": "",
            "attributes": [],
            "methods": []
        }
        self._name = name
        self._bases = bases

        self._set_class(name, bases)

    def __str__(self):
        return (
            f"{self._content['header']}"
            f"{''.join(self._content['class_members'])}\n"
            f"{self._content['constructor']}"
            f"{''.join(sorted(self._content['attributes']))}"
            f"{''.join(sorted(self._content['methods']))}\n\n"
        )

    def __add__(self, other):
        return str(self) + other

    def __radd__(self, other):
        return other + str(self)

    def add_member(self, name, type, ref):
        _type = (self.TYPE_MAP.get(type) or type).format(ref=ref)
        self._content["class_members"].append(
            self.CLASS_MEMBER_TEMPLATE.format(
                member_name=name, type=_type)
        )

    def add_attribute(self, name, type, ref):
        """ add an attribute stub

        Args:
            name (str): attribute name
            type (str): attribute type
            ref (str): attribute type references

        Returns:

        """

        if not self._content["constructor"]:
            raise ValueError(
                "No constructor stub was added. Add an ` __init__` stub via add_method() first."
            )

        _type = (self.TYPE_MAP.get(type) or type).format(ref=ref)

        self._content["attributes"].append(
            self.ATTRIBUTE_TEMPLATE.format(
                attribute_name=name, type=_type
            )
        )

    def add_method(self, name, arguments, keyword_arguments="", return_type="None"):
        """ add a method stub

        Args:
            name (str): method name
            arguments (str): collapsed argument type hints
            keyword_arguments (str): collapses keyword argument type hints
            return_type (str): the expected return type

        Returns:

        """
        if keyword_arguments:
            arguments += ", "

        _method_str = self.METHOD_TEMPLATE.format(
            method_name=name,
            arguments=arguments,
            keyword_arguments=keyword_arguments,
            return_type=return_type,
            ellipsis="{ellipsis}"
        )

        # some special treatment
        if name == "__init__":
            self._content["constructor"] = _method_str.format(ellipsis="")
        else:
            self._content["methods"].append(_method_str.format(ellipsis="..."))

    def _set_class(self, name, bases):
        if bases:
            bases = "(" + ", ".join(list(bases)) + ")"
        else:
            bases = ""
        self._content["header"] = self.CLASS_TEMPLATE.format(class_name=name, class_bases=bases)

    def set_class_bases(self, bases):
        self._set_class(self.name, bases)

    @property
    def name(self):
        return self._name


def get_stubs_from_schemas(include_custom_attributes=False):
    """ generate stubs from available entity schemas

    Returns:
        list: list with StubClassBuiler instances
    """
    stubs = []

    for element in SESSION.schemas:
        stub = StubClassBuilder(name=element["id"])

        # ensure we add the __init__ first, because the stub builder is limited
        # and expects we have that added, to use the proper indentation for the
        # attributes
        for name, _ in element["properties"].items():
            if name == "custom_attributes":
                # skip custom attributes as we will add them individually later
                continue
            items = _.get("items")
            ref = _.get("$ref", "")
            if items:
                ref = _["items"].get("$ref", "")
            if ref:
                ref = "{}".format(ref)
            _type = _.get("type")
            if _type:
                stub.add_member(name=name, type=_type, ref=ref)
            else:
                stub.add_member(name=name, type="any", ref=ref)

        if include_custom_attributes and "custom_attributes" in element["properties"].keys():
            item = SESSION.query("select custom_attributes from {}".format(stub.name)).first()
            if item:
                for key in item["custom_attributes"].keys():
                    stub.add_member(name="custom_" + key, type=_type, ref=ref, cls_attribute=True)

        stubs.append(stub)

    return stubs


def get_extended_entity_stubs(classes, stubs):
    """ extend given stubs for our entity implementations

    Args:
        classes (list): list of classes that need to match the stub names
        stubs (list): list of StubClassBuilder instances

    Returns:
        list: list of StubClassBuilder instances

    """
    stubs_map = {_.name: _ for _ in stubs}
    class_map = dict(classes)

    stubs = []

    for stub_name, stub in stubs_map.items():
        _cls = class_map.get(stub_name)
        if _cls:

            stub.add_method(
                name="__init__",
                arguments="self, *args",
                keyword_arguments="**kwargs",
            )

            for method in inspect.getmembers(
                    _cls,
                    predicate=lambda x: (inspect.isfunction(x) or inspect.ismethod(x)) and not x.__name__.startswith(
                        "__")):

                specs = inspect.getfullargspec(method[1])
                varargs = specs.varargs or ""
                if varargs:
                    varargs = ", *" + varargs

                kwargs = []
                args = specs.args or []
                if specs.defaults:
                    for i, element in enumerate(specs.args[-len(specs.defaults):]):
                        if isinstance(specs.defaults[i], str):
                            kwargs.append("{}=\"{}\"".format(element, specs.defaults[i]))
                        else:
                            kwargs.append("{}={}".format(element, specs.defaults[i]))
                    args = specs.args[:-len(specs.defaults)]

                args = ", ".join(args)
                kwargs = ", ".join(kwargs)

                # always expect the forwarded Query return type
                # in case we are a "by_*" Criteria method
                if re.search(r"^by_", method[0]):
                    return_type = "Query({})".format(stub_name)
                elif re.search(r"^inject$", method[0]):
                    return_type = "Query({})".format(stub_name)
                    args = "self, filter"
                    kwargs = ""
                    varargs = ""
                elif method[0] == "get" or method[0].startswith("get_"):
                    return_type = stub_name
                elif method[0] == "create":
                    return_type = stub_name
                elif method[0] == "create_batch":
                    return_type = stub_name
                elif re.match(r"^(un)?link_.*", method[0]):
                    return_type = stub_name
                elif method[0] == "pre_create":
                    continue
                elif method[0] == "delete":
                    return_type = "None"
                else:
                    return_type = "typing.Any"

                # be stupid and ignore any kind of namespace for base classes
                stub.set_class_bases([_.__name__ for _ in _cls.__bases__])
                stub.add_method(
                    name=method[0],
                    arguments=args + varargs,
                    keyword_arguments=kwargs,
                    return_type=return_type
                )
                # if method is a criterion let's add the negated counterpart
                if re.search(r"^by_", method[0]):
                    stub.add_method(
                        name="not_{}".format(method[0]),
                        arguments=args + varargs,
                        keyword_arguments=kwargs,
                        return_type=return_type
                    )

                # TODO: add overload for get, to add relative terminator

            # TODO improve for base types like TypedContext, Component
            #  that support filtering via Entity
            stub.add_method(
                name="__getitem__",
                arguments="self, item: typing.Union[int, slice, str]",
                keyword_arguments="",
                return_type=stub_name
            )
            stub.add_method(
                name="fetch_attributes",
                arguments="self, projections: typing.List[typing.Union[str, Entity]]",
                keyword_arguments="",
                return_type=stub_name
            )
        stubs.append(stub)

    return stubs


def generate_entitites_stubs():
    """ (re)generate our entities.pyi stub file

    Notes:
        This overrides the stub .pyi file at the expected location. Here we will
        override trixter.entities.entities.pyi with the freshly inspected data.
        Use this whenever the implementation of any Entity subclass changes or
        the ftrack entities schema was modified.

    Returns:

    """
    # TODO: The entities lookup should go away once our project is a proper package.
    #  For now look up the expected package name dynamically based on the location of our script
    #  and make 'entities' importable.
    entities_path = os.path.dirname(os.path.dirname(__file__))
    package_name, package_root = os.path.basename(entities_path), os.path.dirname(entities_path)
    sys.path.insert(0, package_root)
    entities = getattr(__import__(package_name), "entities")
    declarations = getattr(__import__(package_name), "declarations")

    classes = inspect.getmembers(
        entities,
        predicate=lambda entity:
            inspect.isclass(entity)
            and issubclass(entity, (entities.Entity, declarations.ForwardDeclaration))
            and " " not in entity.__name__  # TODO: why do we get members with whitespace in name?
    )

    stubs = STUBS_HEADER
    stubs += "from .base import Entity\n"
    stubs += "from ..query import Query\n\n"

    for stub in sorted(get_extended_entity_stubs(classes, get_stubs_from_schemas()), key=lambda x: x.name):
        stubs += stub

    path = os.path.join(os.path.dirname(entities.__file__), "entities.pyi")
    if os.path.isfile(path):
        os.remove(path)

    LOG.info("Updating stubsfile {}...".format(path))
    make_file(path, default_content=stubs)


if __name__ == '__main__':
    generate_entitites_stubs()



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import importlib
import logging
import os
import traceback
import sys

import wrapt


def _fallback_decorator(fallback_func, attribute):
    """ fallback mechanism for the wrapped function on the associated config entry

    Args:
        fallback_func callable: function to wrap
        attribute: name of the attribute the fallback will be applied to

    Returns:
        callable: function wrapper

    """
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except:
            _LOG.error(
                f"Calling '{attribute}' raised an exception. Falling back to default implementation!",
                exc_info=True
            )
            return fallback_func(*args, **kwargs)

    return wrapper


def _override_configuration():
    """ identifies and parses a user configuration and applies relevant attributes to this module

    """
    _custom_configuration = os.getenv("TRACKTEROID_CONFIGURATION", False)
    # try our best to source a custom configuration
    if _custom_configuration:
        if not os.path.exists(_custom_configuration):
            raise OSError(f"Custom configuration `{_custom_configuration}` doesn't exist.")
        if not os.path.splitext(_custom_configuration)[1] in [".py"]:
            raise AssertionError(f"Custom configuration must `{_custom_configuration}` be a .py file.")

        _LOG.info(f"Custom configuration specified in `{_custom_configuration}`. Trying to load...")
        try:
            spec = importlib.util.spec_from_file_location("custom_configuration", _custom_configuration)
            custom_configuration = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(custom_configuration)
        except:
            raise RuntimeError(
                "Failed to load custom configuration.\n{}".format(
                    "\n".join(traceback.format_exception(*sys.exc_info()))
                )
            )

        relevant_attributes = [_ for _ in globals().keys() if _.isupper() and not _.startswith("_")]
        for attribute in relevant_attributes:
            if hasattr(custom_configuration, attribute):
                _LOG.info(f"Adding '{attribute}' from user configuration.")
                if attribute in _CALLABLES_REQUIRE_FALLBACK:
                    globals()[attribute] = _fallback_decorator(
                        globals()[attribute],
                        attribute
                    )(getattr(custom_configuration, attribute))
                else:
                    globals()[attribute] = getattr(custom_configuration, attribute)


# default entries - ALL upper variable names in here can be set in custom user configuration
############################################################################################

# A functions that gets the current api version string and would return a custom entities
# schema used for relationship resolutions.
RELATIONSHIPS_RESOLVER = lambda api_version: {}

# The primary logger
LOGGING_NAMESPACE = "trackteroid"

# A function that gets the current session and the type name that it requests a
# deletion for and resolves to True or False.
ALLOWED_FOR_DELETION_RESOLVER = lambda session, type_name: True

WARN_ON_INJECT = False
############################################################################################

_LOG = logging.getLogger(f"{LOGGING_NAMESPACE}.configuration")
_CALLABLES_REQUIRE_FALLBACK = [
    "RELATIONSHIPS_RESOLVER",
]

_override_configuration()





#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from .query import (
    Query,
    SCHEMA
)
from .session import SESSION
from .entities import *
from .entities.schematypes import (
    CUSTOM_ATTRIBUTE_TYPE_COMPATIBILITY,
    OBJECT_TYPES,
    ASSET_TYPES,
    TASK_TYPES,
    PROJECT_SCHEMAS
)
from .constants import __VERSION__


#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import re

from arrow.arrow import Arrow
from datetime import datetime
from collections import defaultdict

from ..session import SESSION


class AttributeInfo(object):
    def __init__(self, types=tuple(), array=False, mutable=True):
        self.types = types
        self.array = array
        self.mutable = mutable


class _CustomAttributeTypeCompatibility(object):
    _compatibility_map = {
        "number": (int, float),
        "date": (Arrow, datetime),
        "text": (str),
        "boolean": (bool,),
        "enumerator": (str),
        "dynamic_enumerator": (str),
        "expression": (str)
        }

    def __init__(self):
        self.attributes = {}

    def __getattr__(self, item):
        if not self.attributes:
            configs = SESSION.query("select key, type.name from CustomAttributeConfiguration").all()
            for config in configs:
                self.attributes[config["key"]] = AttributeInfo(
                        types=self._compatibility_map[config["type"]["name"]],
                        )

        if item in self.attributes:
            return self.attributes[item]

        raise KeyError("Custom Attribute '{}' not found.".format(item))


class TypesBase(object):
    projections = ["id", "name"]
    entity_type = None

    def __init__(self):
        self._types = defaultdict(dict)

    def get(self, item, session=SESSION):
        if " " in item:
            item = self._to_camel_case(item)
        session_id = str(session)
        types = self._types[session_id]
        if not types:
            object_types = session.query(
                "select {} from {}".format(", ".join(self.projections), self.entity_type)
            ).all()
            for _type in object_types:
                types[self._to_camel_case(_type["name"])] = _type

        if item in types:
            return types[item]
        elif item == "types":
            return types

        raise KeyError("{} '{}' not found.".format(self.entity_type, item))

    def __getattr__(self, item):
        return self.get(item)

    def __getitem__(self, item):
        return self.get(item)

    @staticmethod
    def _to_camel_case(string):
        return "".join([_.capitalize() for _ in re.findall(r"[a-zA-Z]*\s*?", string)])


class _ObjectTypes(TypesBase):
    entity_type = "ObjectType"


class _AssetTypes(TypesBase):
    entity_type = "AssetType"


class _ProjectSchemas(TypesBase):
    entity_type = "ProjectSchema"


class _TaskTypes(TypesBase):
    entity_type = "Type"


CUSTOM_ATTRIBUTE_TYPE_COMPATIBILITY = _CustomAttributeTypeCompatibility()
OBJECT_TYPES = _ObjectTypes()
ASSET_TYPES = _AssetTypes()
TASK_TYPES = _TaskTypes()
PROJECT_SCHEMAS = _ProjectSchemas()



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import itertools
import importlib
import inspect
import logging
import re


from ..configuration import LOGGING_NAMESPACE


_LOG = logging.getLogger("{}.declarations".format(LOGGING_NAMESPACE))


class ForwardDeclaration:
    pass


class RelationshipDeclaration:
    """ a declaration for a used relationship that needs resolution

    """
    def __init__(self, parent, child):
        self._filter_re = re.compile("\[\w+\]")
        self._chain = []
        self._entities_module = importlib.import_module("..entities", __name__)
        self._chain.extend(
            [
                getattr(self._entities_module, parent.__name__, parent),
                getattr(self._entities_module, child.__name__ if not isinstance(child, str) else child, child)
            ]
        )

    def __getattr__(self, item):
        # TODO: what requests `shape`??
        if not item.startswith("__") and item != "shape":
            self._chain.append(getattr(self._entities_module, item, item))
        return self

    def resolve_path_for(self, entity_type, session, schema):
        relationships = []
        for item in self._chain:
            if item:
                if not isinstance(item, str):
                    item.relationship(session=session, schema=schema)
                    relationship = entity_type.relationship.get(item)
                    if not relationship:
                        _LOG.warning(
                            "Unable to retrieve relationship between target entity type `{}` and entity type `{}`."
                            "You should consider providing via the configurable `RELATIONSHIP_RESOLVER`.".format(
                                entity_type.__class__.__name__,
                                item.__name__
                            )
                        )
                        continue
                    if not isinstance(relationship.relation, list):
                        relationships.append([self._filter_re.sub("", relationship.relation)])
                    else:
                        relationships.append(
                            [self._filter_re.sub("", _) for _ in relationship.relation]
                        )
                    entity_type = item
                else:
                    relationships.append([item])

        return [".".join(_) for _ in itertools.product(*relationships)]


class ForwardDeclareCompare(type):
    def __eq__(self, other):
        if other and inspect.isclass(other) and issubclass(other, ForwardDeclaration):
            return self.__name__ == other.__name__
        else:
            return super(ForwardDeclareCompare, self).__eq__(other)

    def __ne__(self, other):
        return not self == other

    def __getattr__(self, item):
        return RelationshipDeclaration(parent=self, child=item)

    def __hash__(self):
        return hash(self.__class__.__name__)


# TODO: we need to auto-regenerate this when the schema changes
class Appointment(ForwardDeclaration): pass
class Asset(ForwardDeclaration): pass
class AssetBuild(ForwardDeclaration): pass
class AssetCustomAttributeLink(ForwardDeclaration): pass
class AssetCustomAttributeLinkFrom(ForwardDeclaration): pass
class AssetCustomAttributeValue(ForwardDeclaration): pass
class AssetGroup(ForwardDeclaration): pass
class AssetType(ForwardDeclaration): pass
class AssetVersion(ForwardDeclaration): pass
class AssetVersionCustomAttributeLink(ForwardDeclaration): pass
class AssetVersionCustomAttributeLinkFrom(ForwardDeclaration): pass
class AssetVersionCustomAttributeValue(ForwardDeclaration): pass
class AssetVersionLink(ForwardDeclaration): pass
class AssetVersionList(ForwardDeclaration): pass
class AssetVersionStatusChange(ForwardDeclaration): pass
class CalendarEvent(ForwardDeclaration): pass
class CalendarEventResource(ForwardDeclaration): pass
class Component(ForwardDeclaration): pass
class ComponentCustomAttributeLink(ForwardDeclaration): pass
class ComponentCustomAttributeLinkFrom(ForwardDeclaration): pass
class ComponentLocation(ForwardDeclaration): pass
class ContainerComponent(ForwardDeclaration): pass
class Context(ForwardDeclaration): pass
class ContextCustomAttributeLink(ForwardDeclaration): pass
class ContextCustomAttributeLinkFrom(ForwardDeclaration): pass
class ContextCustomAttributeValue(ForwardDeclaration): pass
class Conversation(ForwardDeclaration): pass
class CustomAttributeConfiguration(ForwardDeclaration): pass
class CustomAttributeGroup(ForwardDeclaration): pass
class CustomAttributeLink(ForwardDeclaration): pass
class CustomAttributeLinkConfiguration(ForwardDeclaration): pass
class CustomAttributeLinkFrom(ForwardDeclaration): pass
class CustomAttributeType(ForwardDeclaration): pass
class CustomAttributeValue(ForwardDeclaration): pass
class CustomConfigurationBase(ForwardDeclaration): pass
class Dashboard(ForwardDeclaration): pass
class DashboardResource(ForwardDeclaration): pass
class DashboardWidget(ForwardDeclaration): pass
class Disk(ForwardDeclaration): pass
class EntitySetting(ForwardDeclaration): pass
class Epic(ForwardDeclaration): pass
class Episode(ForwardDeclaration): pass
class Event(ForwardDeclaration): pass
class Feed(ForwardDeclaration): pass
class FileComponent(ForwardDeclaration): pass
class Floor(ForwardDeclaration): pass
class Folder(ForwardDeclaration): pass
class Group(ForwardDeclaration): pass
class GroupCustomAttributeLink(ForwardDeclaration): pass
class GroupCustomAttributeLinkFrom(ForwardDeclaration): pass
class Hardware(ForwardDeclaration): pass
class Job(ForwardDeclaration): pass
class JobComponent(ForwardDeclaration): pass
class License(ForwardDeclaration): pass
class List(ForwardDeclaration): pass
class ListCategory(ForwardDeclaration): pass
class ListCustomAttributeLink(ForwardDeclaration): pass
class ListCustomAttributeLinkFrom(ForwardDeclaration): pass
class ListCustomAttributeValue(ForwardDeclaration): pass
class ListObject(ForwardDeclaration): pass
class ListObjectCustomAttributeValue(ForwardDeclaration): pass
class Location(ForwardDeclaration): pass
class Manager(ForwardDeclaration): pass
class ManagerType(ForwardDeclaration): pass
class Membership(ForwardDeclaration): pass
class Message(ForwardDeclaration): pass
class Metadata(ForwardDeclaration): pass
class Milestone(ForwardDeclaration): pass
class Note(ForwardDeclaration): pass
class NoteCategory(ForwardDeclaration): pass
class NoteComponent(ForwardDeclaration): pass
class NoteLabel(ForwardDeclaration): pass
class NoteLabelLink(ForwardDeclaration): pass
class ObjectType(ForwardDeclaration): pass
class Office(ForwardDeclaration): pass
class Participant(ForwardDeclaration): pass
class Priority(ForwardDeclaration): pass
class Project(ForwardDeclaration): pass
class ProjectSchema(ForwardDeclaration): pass
class ProjectSchemaObjectType(ForwardDeclaration): pass
class ProjectSchemaOverride(ForwardDeclaration): pass
class Recipient(ForwardDeclaration): pass
class Resource(ForwardDeclaration): pass
class Resources(ForwardDeclaration): pass
class ReviewSession(ForwardDeclaration): pass
class ReviewSessionInvitee(ForwardDeclaration): pass
class ReviewSessionObject(ForwardDeclaration): pass
class ReviewSessionObjectStatus(ForwardDeclaration): pass
class Schema(ForwardDeclaration): pass
class SchemaStatus(ForwardDeclaration): pass
class SchemaType(ForwardDeclaration): pass
class Scope(ForwardDeclaration): pass
class Seat(ForwardDeclaration): pass
class SecurityRole(ForwardDeclaration): pass
class Sequence(ForwardDeclaration): pass
class SequenceComponent(ForwardDeclaration): pass
class Setting(ForwardDeclaration): pass
class SettingComponent(ForwardDeclaration): pass
class Shot(ForwardDeclaration): pass
class Sprint(ForwardDeclaration): pass
class State(ForwardDeclaration): pass
class Status(ForwardDeclaration): pass
class StatusChange(ForwardDeclaration): pass
class Task(ForwardDeclaration): pass
class TaskTemplate(ForwardDeclaration): pass
class TaskTemplateItem(ForwardDeclaration): pass
class TaskTypeSchema(ForwardDeclaration): pass
class TaskTypeSchemaType(ForwardDeclaration): pass
class TestTest(ForwardDeclaration): pass
class Timelog(ForwardDeclaration): pass
class Timer(ForwardDeclaration): pass
class Type(ForwardDeclaration): pass
class TypedContext(ForwardDeclaration): pass
class TypedContextLink(ForwardDeclaration): pass
class TypedContextList(ForwardDeclaration): pass
class TypedContextStatusChange(ForwardDeclaration): pass
class User(ForwardDeclaration): pass
class UserApplicationState(ForwardDeclaration): pass
class UserCustomAttributeLink(ForwardDeclaration): pass
class UserCustomAttributeLinkFrom(ForwardDeclaration): pass
class UserCustomAttributeValue(ForwardDeclaration): pass
class UserSecurityRole(ForwardDeclaration): pass
class UserSecurityRoleProject(ForwardDeclaration): pass
class UserType(ForwardDeclaration): pass
class WorkflowSchema(ForwardDeclaration): pass
class WorkflowSchemaStatus(ForwardDeclaration): pass


from typing import *
from ..session import Session

T = TypeVar('T')

class Query(Generic[T]):
    _known_order: ClassVar[List[str]] = ["ascending", "descending"]

    def __new__(cls: Type[Query[T]], _cls: T, session: Optional[Session], schema: Optional[dict]) -> T: ...


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import time

project = 'Trackteroid'
copyright = f'2023-{time.strftime("%Y")} Trixter Film GmbH'
author = 'Dennis Weil, Rico Koschmitzky'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
# Add the following lines at the top of the file
import myst_parser
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

extensions = [
    'myst_parser',
    'sphinx_togglebutton',
]
myst_enable_extensions = [
    'colon_fence',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

html_favicon = "_static/favicon.ico"
html_logo = '_static/logo_white.svg'

suppress_warnings = [
    'myst.xref_missing',
    'myst.header'
]


#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# TODO: Move this into query.py
def build_partial_query(target, arguments, attribute):
    """

    Args:
        target (TargetRelation): target relation data (relation + collection).
        arguments (tuple): attribute values/patterns to match in the query.
        attribute (str): attribute chain.

    Returns:
        str: partial query expression.

    """
    def _build_for_relation(relation, attribute):
        pattern_match_expressions = []
        non_pattern_elements = []
        for element in arguments:
            if isinstance(element, str) and "%" in element:
                pattern_match_expressions.append(
                        "{} like \"{}\"".format(attribute, element)
                )
            else:
                non_pattern_elements.append("\"{}\"".format(element))

        non_pattern_expressions = []

        # handle known id shortcuts like parent.id -> parent_id, version.id -> version_id
        if attribute == "id" and target.relation and not target.collection:
            relation_tokens = relation.split(".")
            if len(relation_tokens) > 1:
                attribute = "{}_id".format(relation_tokens[-1])
                relation = ".".join(relation_tokens[:-1])
            else:
                attribute = "{}_id".format(relation_tokens[0])
                relation = ""

        if len(non_pattern_elements) == 1:
            non_pattern_expressions.append(
                "{} is {}".format(attribute, non_pattern_elements[0])
            )
        elif non_pattern_elements:
            non_pattern_expressions.append(
                "{} in ({})".format(attribute, ",".join(non_pattern_elements))
            )

        # Attribute expression
        query = "({})".format(" or ".join(non_pattern_expressions + pattern_match_expressions))

        if relation:
            # Relation expression
            query = "{} {} {}".format(relation, "any" if target.collection else "has", query)

        return query

    relations = [target.relation] if not isinstance(target.relation, list) else target.relation
    if len(relations) > 1:
        start, end = "(", ")"
    else:
        start, end = "", ""

    return start + " or ".join(_build_for_relation(relation, attribute) for relation in relations) + end



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Trixter GmbH
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  1. Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#  3. Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import dbm
import contextlib
import json
import warnings
import inspect
import importlib

import ftrack_api

from ftrack_api.symbol import NOT_SET

from ftrack_api.cache import (
    SerialisedCache,
    FileCache
)
from ftrack_api.operation import (
    CreateEntityOperation,
    UpdateEntityOperation,
    Operations
)

from .entities.relationships_parser import RelationshipsParser


_PARSED_RELATIONSHIPS_CACHE = {}


class Session(object):
    """ Delegates all attributes and methods to the ftrack session instance

    Allows us to have our own Session interface for enhanced interaction.
    """

    # exclude members that we don't want to delegate
    MEMBERS = [
        "_session",
        "_session_arguments",
        "_terminated_queries",  # this is our terminated query cache
        "reuse_query_results",
    ]

    def __init__(self, auto_populate=False, auto_connect_event_hub=False, **kwargs):
        # Collect arguments to hand over to the native ftrack_api.Session
        local_args = dict(locals())
        valid_args = [_ for _ in inspect.getfullargspec(ftrack_api.Session.__init__).args if _ not in ('self', 'kwargs')]
        session_arguments = {}
        for arg_key, arg_val in local_args.items():
            if arg_key in valid_args:
                session_arguments[arg_key] = arg_val
        session_arguments.update(kwargs)

        self._session = ftrack_api.Session(**session_arguments)

        # monkey patch ftrack session to find the way back from it to here
        self._session.delegate = self
        self._session_arguments = session_arguments
        self._terminated_queries = {}

        self.reuse_query_results = kwargs.get("reuse_query_results", False)

    def __eq__(self, other):
        if isinstance(other, ftrack_api.Session):
            return other == self._session
        elif isinstance(other, Session):
            return other._session == self._session
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def parsed_relationships(self):
        if not self._session.server_url in _PARSED_RELATIONSHIPS_CACHE:
            relationship_parser = RelationshipsParser(ftrack_session=self._session)
            # types such as "Context" are ambiguous and cannot be used to extract direct relationships
            # example structure - Seq -> Folder -> Shot
            # in this case, establishing a relationship from Seq to Shot is impossible as would any other
            # that went through "parent" or "children" for example
            relationship_parser.exclude_types = {"Context"}
            relationship_parser.parse_session_schemas()
            # attributes that group things cannot be used to extract relationships
            # as an example, "descendants" lists all children and children of children and so on,
            # so they are all flattened with their relationships removed
            relationship_parser.attributes_blacklist = {"descendants", "ancestors", "status_changes"}
            relationship_parser.extract_all_entity_relations()
            _PARSED_RELATIONSHIPS_CACHE[self._session.server_url] = relationship_parser

        return _PARSED_RELATIONSHIPS_CACHE[self._session.server_url]

    def get_cached_collections(self):
        typenames = sorted([str(k) for k, v in self._build_entity_type_classes(self.schemas).items()])
        type_module = importlib.import_module("..entities", __name__)

        type_map = {}
        for name in typenames:
            collection = getattr(type_module, "EntityCollection")
            _type = getattr(type_module, name)
            if _type and issubclass(_type, getattr(type_module, "Entity")):
                cached_entities = [
                    _type.from_entity_type(name=_type.__name__, ftrack_entity=_) for _ in self._local_cache.values()
                    if _.__class__.__name__ == name
                ]
                entitycollection = collection._make_empty(_type, self)
                type_map[_type] = entitycollection.from_entities(cached_entities)

        return type_map

    def reconnect(self, **kwargs):
        """ Closes the active session and connects a new session with the
        option to update session arguments.

        Args:
            kwargs: optional overrides for existing session arguments.
        """
        session_attributes = {}
        valid_args = [_ for _ in inspect.getfullargspec(ftrack_api.Session.__init__).args if _ not in ('self', 'kwargs')]
        for arg in valid_args:
            if hasattr(self._session, arg):
                session_attributes[arg] = getattr(self._session, arg)

        self._session.close()
        self._session_arguments.update(session_attributes)
        self._session_arguments.update(kwargs)
        self._session = ftrack_api.Session(**self._session_arguments)
        # Some of the attributes propagated as arguments are cache-related, therefore we
        # need to clear the cache in order to make sure it's empty for the new session.
        self._session.cache.clear()

        self._terminated_queries = {}

    def close(self):
        """Closes the session"""
        self._session.close()
        self._terminated_queries = {}

    def reconnect_and_commit(self, initial_operations_path, **kwargs):
        """
        Given a path to a dumped database [.dbm] this method will apply
        the stored operations to the database.

        The session is reconnected to start from a clean cache and
        an empty operations stack. Then the local cache and operations
        will be updated with the contents of the dumped cache.
        To sync the ftrack database with the local cache a commit is
        issued.

        Args:
            initial_operations_path ():
            kwargs: optional overrides for existing session arguments.

        Returns:

        """
        self.reconnect(**kwargs)

        operations = Operations()

        try:
            _db = dbm.open(initial_operations_path, "r")
            _db.close()
        except dbm.error:
            raise OSError("Unable to open Database `{}`".format(initial_operations_path))

        initial_cache = SerialisedCache(
            FileCache(initial_operations_path),
            encode=self._session.encode,
            decode=self._session.decode
        )
        for key in initial_cache.keys():
            if key == "__operations__":
                initial_operations = json.loads(initial_cache.get(key))

                for op in initial_operations:
                    if op["operation"] == "create":
                        del op["operation"]
                        operation = CreateEntityOperation(**op)
                    elif op["operation"] == "update":
                        if op["old_value"] in initial_cache.keys():
                            op["old_value"] = initial_cache.get(str(op["old_value"]))
                            if op["old_value"] == "NOT SET":
                                op["old_value"] = NOT_SET
                        if op["new_value"] in initial_cache.keys():
                            op["new_value"] = initial_cache.get(str(op["new_value"]))
                            if op["new_value"] == "NOT SET":
                                op["new_value"] = NOT_SET
                        del op["operation"]
                        operation = UpdateEntityOperation(**op)
                    else:
                        raise KeyError("Unknown Operation `{}`.".format(op["operation"]))

                    operations.push(operation)
            else:
                self._local_cache.set(key, initial_cache.get(key))

        self._session.recorded_operations = operations
        self._session.commit()

    @staticmethod
    def _serialise_value(value):
        if value == NOT_SET:
            return "NOT SET"
        elif hasattr(value, "entity_type"):
            return "('{}', ['{}'])".format(
                value.entity_type,
                value["id"]
            )
        else:
            return value

    @contextlib.contextmanager
    def reusing_query_results(self):
        """Temporarily enable query cache"""
        backup = self.reuse_query_results
        self.reuse_query_results = True
        try:
            yield
        finally:
            self.reuse_query_results = backup
            
    @contextlib.contextmanager
    def deferred_operations(self, filepath, clear=True):

        file_cache = FileCache(filepath)
        serialised_cache = SerialisedCache(
            file_cache,
            encode=self.encode,
            decode=self.decode
        )

        if clear:
            serialised_cache.clear()

        # add the filecache temporarily
        self.cache.caches.append(serialised_cache)
        cache_index = len(self.cache.caches) - 1

        # store all previously recorded operations
        _previous_operations = self.recorded_operations

        # clear operations temporarily
        deferred = Operations()
        self.recorded_operations = deferred

        yield

        # sync cache
        with file_cache._database() as database:
            for key, value in database.items():
                entity_data = json.loads(value)
                for attr, attr_value in entity_data.items():
                    if isinstance(attr_value, dict):
                        entity_type = attr_value.get("__entity_type__")
                        if entity_type:
                            # if this is an entity we check whether the linked entity is available
                            # in our filecache
                            # TODO: eventually we should check for the actual primary key here
                            #  for our current usecase it seems ok to blindly use the id
                            cache_key = "('{}', ['{}'])".format(entity_type, attr_value["id"])
                            if cache_key not in serialised_cache.keys():
                                serialised_cache.set(cache_key, self._local_cache.get(cache_key))

            # serialise operations and store them in the database
            ops = []
            for op in self.recorded_operations:
                if isinstance(op, CreateEntityOperation):
                    ops.append(
                        {
                            "operation": "create",
                            "entity_data": op.entity_data,
                            "entity_type": op.entity_type,
                            "entity_key": op.entity_key
                        }
                    )
                elif isinstance(op, UpdateEntityOperation):
                    ops.append(
                        {
                            "operation": "update",
                            "entity_type": op.entity_type,
                            "entity_key": op.entity_key,
                            "attribute_name": op.attribute_name,
                            "old_value": self._serialise_value(op.old_value),
                            "new_value": self._serialise_value(op.new_value)
                        }
                    )

            serialised_cache.set("__operations__", json.dumps(ops))

        # remove temporary filecache
        self.cache.caches.pop()

        # load previous operations back in the stack
        self.recorded_operations = _previous_operations

    def __getattribute__(self, item):
        # delegating access to those attributes is neccessary to support the
        # temporary attribute value decorator
        if item in ("__dict__", "__slots__"):
            return self.__getattr__(item)
        else:
            return super(Session, self).__getattribute__(item)

    def __getattr__(self, key):
        if hasattr(self._session, key):
            return getattr(self._session, key)
        else:
            raise AttributeError("Attribute '{}' does not exist.".format(key))

    def __setattr__(self, key, value):
        if key in self.MEMBERS:
            super(Session, self).__setattr__(key, value)
            setattr(self._session, key, value)
        elif hasattr(self, key):
            setattr(self._session, key, value)
        else:
            raise AttributeError("Attribute '{}' does not exist.".format(key))

    def __delattr__(self, item):
        if item in self.MEMBERS:
            super(Session, self).__delattr__(item)
            delattr(self._session, item)
        elif hasattr(self, item):
            delattr(self._session, item)
        else:
            raise AttributeError("Attribute '{}' does not exist.".format(item))


class _SessionDelegate(Session):
    def __init__(self, *args, **kwargs):
        warnings.warn('"_SessionDelegate" class got renamed to "Session"',
                      DeprecationWarning)
        super(_SessionDelegate, self).__init__(*args, **kwargs)


# main session
# TODO: defer connection
SESSION = Session()
SESSION.auto_populate = False
